/* 4 stack processor emulator */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "4stack.h"

int MEMSIZE =0x1000000;    /* 16 MB should be enough for simple tests :-) */
int FRAMEBUF=0x0800000;    /* therefrom 8 MB frame buffer - wow! */

#define IP(addr)  *(opcode *)((int)ip + (int)image + addr)

int LINES=
#ifndef MSDOS
	24;
#else
	25;
#endif

#ifdef WITH_SCREEN
extern int w, h;
#endif

ull dtime(void)
{
   struct timeval time1;
   struct timezone zone1;
   dcell time;
	
   gettimeofday(&time1,&zone1);
   
   time.hl.hi=time1.tv_sec;
   time.hl.lo=(int)(((ull)time1.tv_usec<<32)/1000000);
   
   return time.d;
}

int loader(void *image, FILE* fileid)
{
  struct{int addr, size;} buf;
  char magic[9];

  magic[8]='\0';

  do
  {
    if(fread(magic,1,8,fileid) < 8) return 1;
#ifdef DEBUG
    printf("Magic found: %s\n",magic);
#endif
  }
  while(memcmp(magic,"4stack00",8));

  while(fread(&buf,4,2,fileid) == 2)
  {
#ifdef DEBUG
    printf("Reading block at %08x with size %08x\n",buf.addr,buf.size);
#endif
    switch(buf.addr & 7)
      {
      case 0:
      case 1:
	fread(image+(buf.addr & -8),buf.size,1,fileid); break;
      case 2:
	bzero(image+(buf.addr & -8),buf.size); break;
      case 3:
	fseek(fileid,buf.size,1); break;
      default:
      }
  }
  return 0;
}

int main(int argc, char** argv)
{
  opcode *ip=(opcode *)0;
  void *image;
  ull starttime, endtime;  

  cell * sp0 = (cell *) 0xE780;
  cell * sp1 = (cell *) 0xEF80;
  cell * sp2 = (cell *) 0xF780;
  cell * sp3 = (cell *) 0xFF80;

#ifndef MSDOS
  install_signal_handlers();
  {
    char *linestr;
    int lines=LINES;

    linestr=getenv("LINES");
    if(linestr) lines=atoi(linestr);
    if(lines)   LINES=lines;
  }
#endif

  if(!(image = (void *)malloc(MEMSIZE)))
     {
	fprintf(stderr,"Unable to allocate virtual memory");
	exit(2);
     }

  FRAMEBUF=(int)(image + (int)FRAMEBUF);

  if(loader(image,fopen(argc > 1 ? argc--, *(++argv) : "test.4o","rb")))
    return 1;

  init(argc, argv, 0);

  ip=(opcode *)8;
   
#ifndef MSDOS
  prep_terminal();
#endif

  starttime = cpustate[2] = dtime();

  emulate(image,(opcode*)ip,sp0,sp1,sp2,sp3);

  endtime = dtime();

#ifndef MSDOS
  deprep_terminal();
#	ifdef WITH_SCREEN
  if(dpy)
    end_graphics();
#	endif
#endif

  printf("%i ops in %i cycles and %f seconds (%f real, %d ips)\n",
	 (int)cpustate[0],(int)cpustate[1],
	 ((double)(cpustate[2]-starttime)/4294967296.0),
	 ((double)(endtime-starttime)/4294967296.0),
	 (int)((double)(cpustate[0])/((double)(endtime-starttime)/4294967296.0)));

  return 0;
}
