typedef unsigned  ucell;
typedef unsigned long long  ull;
typedef long long ll;
typedef int cell;

#define PTR(ptr)   *(typeof(ptr)) ((void *)(image) + (int)(ptr))

#ifndef BIGENDIAN
#  ifdef i386
#    define BIGENDIAN 0
#  else
#    define BIGENDIAN 1
#  endif
#endif

#if BIGENDIAN
#   define THI(x)      ((ucell *)&x)[0]
#   define TLO(x)      ((ucell *)&x)[1]
#else
#   define THI(x)      ((ucell *)&x)[1]
#   define TLO(x)      ((ucell *)&x)[0]
#endif
/*
#if BIGENDIAN
#   define HI(x)      ((ucell *)&x)[0]
#   define LO(x)      ((ucell *)&x)[1]
#else
#   define HI(x)      ((ucell *)&x)[1]
#   define LO(x)      ((ucell *)&x)[0]
#endif
#if BIGENDIAN
#   define FHI(x)      ((ucell *)&x)[0]
#   define FLO(x)      ((ucell *)&x)[1]
#else
#   define FHI(x)      ((ucell *)&x)[1]
#   define FLO(x)      ((ucell *)&x)[0]
#endif
*/

#if BIGENDIAN
typedef union
{
   ull d;
   double df;
   struct
     {
	ucell hi, lo;
     } hl;
} dcell;
#else
typedef union
{
   ull d;
   double df;
   struct
     {
	ucell lo, hi;
     } hl;
} dcell;
#endif

#define HI(x)	({ dcell _d; _d.d=x; _d.hl.hi; })
#define LO(x)	({ dcell _d; _d.d=x; _d.hl.lo; })

#define FHI(x)	({ dcell _d; _d.df=x; _d.hl.hi; })
#define FLO(x)	({ dcell _d; _d.df=x; _d.hl.lo; })

/* #define HI(x)      ((ucell) ((x) >> 32)) */

/* typedef ull opcode; */
#if BIGENDIAN
typedef union
{
   struct
     {
	ucell alu0:10, alu1:10, alu2:10, alu3a:2;
	ucell alu3b:8, data0:10, data1:10, cu0:1, cu1:1, typ:2;
     } cont;
   struct
     {
	ucell alu0:10, alu1:10, alu2:10, alu3a:2;
	ucell alu3b:8, cc0:5, cc1:5, cc2:5, cc3:5, subtype:2, typ:2;
     } csetup;
   struct
     {
	ucell alu0:10, alu1:10, alu2:10, alu3a:2;
	ucell alu3b:8, mask:5, cc:5; int offset:11; ucell likelihood:1, typ:2;
     } branch;
   struct
     {
	ucell alu0:10, alu1:10, alu2:10, abs:1, cj:1;
	ucell offset:29, far:1, typ:2;
     } jump;
   ull data;
} opcode;
#else
typedef union
{
   struct
     {
	ucell typ:2, cu1:1, cu0:1, data1:10, data0:10, alu3b:8;
	ucell alu3a:2, alu2:10, alu1:10, alu0:10;
     } cont;
   struct
     {
	ucell typ:2, subtype:2, cc3:5, cc2:5, cc1:5, cc0:5, alu3b:8;
	ucell alu3a:2, alu2:10, alu1:10, alu0:10;
     } csetup;
   struct
     {
	ucell typ:2, likelihood:1; int offset:11; ucell cc:5, mask:5, alu3b:8;
	ucell alu3a:2, alu2:10, alu1:10, alu0:10;
     } branch;
   struct
     {
	ucell typ:2, far:1, offset:29;
	ucell cj:1, abs:1, alu2:10, alu1:10, alu0:10;
     } jump;
   ull data;
} opcode;
#endif

#ifdef undefined
#  define ALU0(x)    (ucell) (((int)HI(x) >> 22) & 0x3FF)
#  define ALU1(x)    (ucell) (((int)HI(x) >> 12) & 0x3FF)
#  define ALU2(x)    (ucell) (((int)HI(x) >> 02) & 0x3FF)
#  define ALU3(x)    (ucell) ((ucell)((x) >> 24) & 0x3FF)
#  define DATA0(x)   (ucell) (((int)(x) >> 14) & 0x3FF)
#  define DATA1(x)   (ucell) (((int)(x) >> 04) & 0x3FF)
#  define CU0(x)     (ucell) (((int)(x) >> 03) & 0x1)
#  define CU1(x)     (ucell) (((int)(x) >> 02) & 0x1)
#  define TYP(x)     ((ucell) (x)         & 0x3)

#  define COND0(x)   (ucell) (((int)(x) >> 19) & 0x1F)
#  define COND1(x)   (ucell) (((int)(x) >> 14) & 0x1F)
#  define COND2(x)   (ucell) (((int)(x) >>  9) & 0x1F)
#  define COND3(x)   (ucell) (((int)(x) >>  4) & 0x1F)
#  define SUBTYPE(x) (ucell) (((int)(x) >>  2) & 0x3)

#  define MASK(x)    (ucell) (((int)(x) >> 19) & 0x1F)
#  define COND(x)    (ucell) (((int)(x) >> 14) & 0x1F)
#  define BROFF(x)   ((int) (((int)(x) >>  3) & 0x3FF) - (int) (((int)(x) >> 3) & 0x400))
#  define DOOFF(x)   ((int) (((int)(x) >>  3) & 0x3FF) - (int) (((int)(x) >> 3) & 0x400))

#  define COFF(x)    ((int)  (LO(x) >> 3) & 0x1FFFFFFF)
#  define CJTYP(x)   (int) (HI(x) & 0x1)
#  define CJABS(x)   (int) (HI(x) & 0x2)
#  define CJLEN(x)   (int) (LO(x) & 0x4)

#  define CFOFF(x)          (LO(x) & -0x8)
#else
#  define ALU0(x)    (x.cont.alu0)
#  define ALU1(x)    (x.cont.alu1)
#  define ALU2(x)    (x.cont.alu2)
#  define ALU3(x)    ((x.cont.alu3a << 8)|(x.cont.alu3b))
#  define DATA0(x)   (x.cont.data0)
#  define DATA1(x)   (x.cont.data1)
#  define CU0(x)     (x.cont.cu0)
#  define CU1(x)     (x.cont.cu1)
#  define TYP(x)     (x.cont.typ)

#  define COND0(x)   (x.csetup.cc0)
#  define COND1(x)   (x.csetup.cc1)
#  define COND2(x)   (x.csetup.cc2)
#  define COND3(x)   (x.csetup.cc3)
#  define SUBTYPE(x) (x.csetup.subtype)

#  define MASK(x)    (x.branch.mask)
#  define COND(x)    (x.branch.cc)
#  define BROFF(x)   (x.branch.offset)
#  define DOOFF(x)   (x.branch.offset)

#  define COFF(x)    (x.jump.offset)
#  define CJTYP(x)   (x.jump.cj)
#  define CJABS(x)   (x.jump.abs)
#  define CJLEN(x)   (x.jump.far)

#  define CFOFF(x)   (LO(x.data) & -0x8)
#endif

#define ALUTYP(x)         (x > 0x1FF)

#define ALUOP(x)          ((x >> 5) & 0x0F)
#define ALUCLASS(x)       ((x >> 7) & 0x03)
#define ALUMOD(x)         (x & 0x1F)

#define IMOP(x)           ((x >> 8) & 0x1)
#define IMVAL(x)          (x & 0xFF)

/* important instructions */

#define IP_FETCH   0x1D6
#define NOOP       0x000

/* flag specifiers */

#define EXTDEBUG    0x10000000
#define SUPERVISOR -0x80000000
#define INTDEBUG    0x40000000
#define IPADDR      0x20000000

/* function declarations */

extern ll cpustate[256];
extern int dpy;
extern int mag;

void emulate(ull *image, opcode *ip,
	     cell *sp0, cell *sp1, cell *sp2, cell *sp3);

ull ioin(ucell address, ull * image);
void ioout(ucell address, ull data, ull * image);
extern ull io_q;

unsigned char getkey(FILE *stream);
void disass(opcode op);
long key_avail (FILE *stream);
void check_event(int what_to_do);
void init(int argc, char** argv, int flag);
void resize();

#ifndef MSDOS
void deprep_terminal ();
void prep_terminal ();
void install_signal_handlers ();
void end_graphics ();
#endif
