/* 4 stack machine disassembler */

#include <stdio.h>
#include "4stack.h"

const char *stack[32] =
{
   "#0","#-1","#max","#min",   "c0","c1","c2","c3",
   "s0p","s1p","s2p","s3p",    "s4","s5","s6","s7",
   "0s0","0s1","0s2","0s3",    "1s0","1s1","1s2","1s3",
   "2s0","2s1","2s2","2s3",    "3s0","3s1","3s2","3s3"
};

char scratch[20];

const char *T2[32] =
{
   "pin s0"  ,"pin s1",  "pin s2",  "pin s3",
   "pin s4"  ,"pin s5",  "pin s6",  "pin s7",

   "pick s0p","pick s1p","pick s2p","pick s3p",
   "pick s4" ,"pick s5" ,"pick s6" ,"pick s7",
   "pick 0s0","pick 0s1","pick 0s2","pick 0s3",
   "pick 1s0","pick 1s1","pick 1s2","pick 1s3",
   "pick 2s0","pick 2s1","pick 2s2","pick 2s3",
   "pick 3s0","pick 3s1","pick 3s2","pick 3s3"
};

const char * T3[32] =
{
   "mul@",    "mul<@",    "mulr@",    "mulr<@",
   "-mul@",   "-mul<@",   "-mulr@",   "-mulr<@",
   "mul@+",   "mul<@+",   "mulr@+",   "mulr<@+",
   "-mul@+",  "-mul<@+",  "-mulr@+",  "-mulr<@+",
   "t",  "0=",   "0<",   "ov",  "u<",   "u>",   "<",   ">",
   "f",  "0<>",  "0>=",  "no",  "u>=",  "u<=",  ">=",  "<="
};

const char * T4[32] =
{
   "asr",  "lsr",    "ror",   "rorc",   "asl",   "lsl",   "rol",   "rolc",
   "ff1",  "popc",   "lob",   "loh",    "extb",  "exth",  "hib",   "hih",
   "sp@",  "loops@", "loope@","ip@",    "sr@",   "cm@",   "index@","flatch@",
   "sp!",  "loops!", "loope!","ip!",    "sr!",   "cm!",   "index!","flatch!"
};

const char * T5[32]=
{
   "fadd",     "fsub",     "fmul",     "fnmul",
   "faddadd",  "faddsub",  "fmuladd",  "fmulsub",
   "fi2f",     "fni2f",    "fadd@",    "fmul@",
   "fs2d2",    "fd2s",     "fxtract",  "fiscale",

/* "ext",      "extu",     "mak",      "clr", */
   "bfu",      "bfs",      "???",      "???",
   "???",      "???",      "???",      "???",
   "px1",      "px2",      "px4",      "px8",
   "pp1",      "pp2",      "pp4",      "pp8"
};

const char * aluname[16] =
{
   "or ",     "add ",   "addc ",   "mul ", 
   "and ",    "sub ",   "subc ",   "umul ",
   "xor ",    "subr ",  "subcr ",  "pass ",
   "",        "",       "",        ""
};

const char ** alumods[16] =
{
   stack, stack, stack, stack,
   stack, stack, stack, stack,
   stack, stack, stack, stack,
   T2,    T3,    T4,    T5
};

void aludis(ucell alu)
{
   const char* rest = "         ";
   
   if(ALUTYP(alu))
     {
	if(alu & 0x100)
	   sprintf(scratch,"$%02x #<", alu & 0x0FF);
	else if(alu & 0x80)
	   sprintf(scratch,"-$%02x #",-(int)(signed char)(alu));
	else
	   sprintf(scratch,"$%02x #", alu & 0x0FF);
     }
   else
     {
       if(ALUOP(alu))
	 sprintf(scratch,"%s%s",
		 aluname[ALUOP(alu)],
		 alumods[ALUOP(alu)][ALUMOD(alu)]);
       else
	 sprintf(scratch,"nop");
     }
   fprintf(stderr,"%s%s",scratch,rest+strlen(scratch));
}

const char * msize[4] = { "b","h","","2" };
const char * mstack[8] = 
{
   "0", "0&2", "2", "2&0",
   "1", "1&3", "3", "3&1"
};
const char * mmode[8] =
{
   "", "+N" , "N+" , "+N+",
   "", "+s0", "s0+", "+s0+"
};
const char * mxmode[4] =
{
   "ip", "s0b", "ip+s0", "s0l"
};
const char * get_ops[4] =
{
  "get", "set", "getd", "setd"
};
const char * cmu_ops[16] =
{
   "ccheck", "cclr",  "cstore", "cflush",
   "???",    "cload", "calloc", "cxlock",
   "mccheck", "mcget", "mcset", "mcpart",
   "mdcheck", "mdget", "mdset", "mdpart"
};
const char * div_ops[16] =
{
  "inb %x:", "inh %x:", "in %x:", "ind %x:",
  "outb %x: R%x", "outh %x: R%x", "out %x: R%x", "outd %x: R%x",
  "", "\02%%s", "geta", "\01getd%%s",
  "%x: =c%x", "\03%%s", "seta", "\01set%%s"
};
const char * div_ops0[2] = { "rn", "mf" };
const char * div_ops1[4] = { "getc", "stop", "stop1", "halt" };
const char * div_ops2[4] = { "setc", "restart", "restart1", "halt" };
const char ** div_op[3] = { div_ops0, div_ops1, div_ops2 };

void datdis(int dat, int cu, int n)
{
   const char *rest="               ";
   char scratch1[20];
   
   if(cu)
     {
	sprintf(scratch,"%s%s %s: ",
	       ((dat & 0x200) ? "st" : "ld"),
	       msize[(dat >> 5) & 3],
	       mstack[((dat >> 7) & 3) + (n << 2)]);
	if((dat & 0x1C) != 0x10)
	   sprintf(scratch+strlen(scratch),
		   "R%x %s",dat & 3,mmode[(dat >> 2) & 7]);
	else
	   sprintf(scratch+strlen(scratch),
		   "%s",mxmode[dat & 3]);
     }
   else
     {
	if(dat & 0x200)
	  {
	     switch((dat >> 6) & 3)
	       {
		  case 0:  if(dat & 0x10)
		              sprintf(scratch,"R%x= R%x %x: %cs0",
				     (dat >> 2) & 3, dat & 3,
				     ((dat >> 7) & 2) + n,
                                     (dat & 0x20) ? '-' : '+');
			   else
		              sprintf(scratch,"R%x= R%x %s",
				     (dat >> 2) & 3, dat & 3,
				     (dat & 0x100) ? (dat & 0x20) ? "-N" : "+N" : "");
			   break;
		  case 1:  sprintf(scratch,"%s %x: %c%x",
				   get_ops[(dat >> 4) & 3],
				   ((dat >> 7) & 2) + n,
				   "RNMF"[((dat >> 2) & 3)],
				   dat & 3); break;
		  case 2:  sprintf(scratch,"%s %x: R%x",
				   cmu_ops[(dat >> 2) & 0xF],
				   ((dat >> 7) & 2) + n,
				   dat & 3); break;
		  case 3:  sprintf(scratch1,
				   div_ops[(dat >> 2) & 0xF],
				   ((dat >> 7) & 2) + n,
				   dat & 3);
		    sprintf(scratch, scratch1 + (scratch1[0] <= '\03'),
			    (scratch1[0] <= '\03') ? div_op[scratch1[0]-1]
			    [scratch1[0]=='\01' ? (dat >> 8) & 1 : dat & 3] : "");
		    break;
	       }
	  }
	else if(dat & 0x100)
	   sprintf(scratch,"$%02x #<", dat & 0x0FF);
	else if(dat & 0x80)
	   sprintf(scratch,"-$%02x #",-(int)(signed char)(dat));
	else
	   sprintf(scratch,"$%02x #", dat & 0x0FF);
     }
   fprintf(stderr,"%s%s",scratch,rest+strlen(scratch));
}

void pcond(int cond)
{
   fprintf(stderr,"%c%s ",(cond & 0x10) ? '?' : ':',T3[cond | 0x10]);
}

void pmask(int mask)
{
   char sep=(mask & 0x10) ? '&' : '|';
   int i,j;
   
   for(i=8, j=0; i; i>>=1, j++)
     if(mask & i)
       fprintf(stderr,"%x%c",j,(mask & (i-1)) ? sep : ' ');
}

void pdist(int n)
{
   if(n<0)
      fprintf(stderr,"-$%x +IP",-n);
   else
      fprintf(stderr,"$%x +IP",n);
}

void aludis4(opcode op)
{
   aludis(ALU0(op));
   aludis(ALU1(op));
   aludis(ALU2(op));
   aludis(ALU3(op));
}

void disass(opcode op)
{
   char * rest =
#ifdef MSDOS
   "\n"
#else
   ""
#endif
   ;

   switch(TYP(op))
     {
	case  0:
	  aludis4(op);
	  datdis(DATA0(op),CU0(op), 0);
	  datdis(DATA1(op),CU1(op), 1);
	  break;
	case  1:
	  aludis4(op);
	  pcond(COND0(op));
	  pcond(COND1(op));
	  pcond(COND2(op));
	  pcond(COND3(op)); break;
	case  2:
	  aludis4(op);
	  if(COND(op) | MASK(op))
	    {
	       fprintf(stderr,"br ");
	       pmask(MASK(op));
	       pcond(COND(op));
	       pdist(BROFF(op));
	    }
	  else
	    {
	       fprintf(stderr,"do ");
	       pdist(DOOFF(op));
	    }
	  break;
	case  3:
	  aludis(ALU0(op));
	  aludis(ALU1(op));
	  aludis(ALU2(op));
	  if(CJTYP(op))
	     fprintf(stderr,"jmp ");
	  else
	     fprintf(stderr,"call ");
	  pdist(COFF(op));
	  break;
     }

   fprintf(stderr,rest);
}
