/* 4 stack processor emulator */

#include <stdio.h>
#include <signal.h>
#include "4stack.h"

#if BIGENDIAN
#   define SWITCHEND(f)     (f & 2)
#   define HIH              0
#   define LOH              1
#   define PART(n,m)        (n)
#else
#   define SWITCHEND(f)     (!(f & 2))
#   define HIH              1
#   define LOH              0
#   define PART(n,m)        ((n)^(m))
#endif

#define STACKSIDE(d)      (d & 0x100)

#define NO_X              0x08
#define EXECQ(sr)         !((sr) & NO_X)
#define NEXECQ(sr)        ((sr) & NO_X)

#define SPT(x)            *(cell*)(((cell)(x))^(!BIGENDIAN << 2))

extern int LINES;
extern int pixels, w, h;

typedef struct
{
   ucell r, n, m, f;
} memreg;

typedef void * Label;

#ifdef i386
#  if (__GNUC__==2 && defined(__GNUC_MINOR__) && __GNUC_MINOR__<7)
    extern ll mulsign(int, int);
    extern ll mulunsign(int, int);
#  else
#    define mulunsign(a, b) (((ll)(ucell)(a)) * (ucell)(b))
#    define mulsign(a, b) (((ll)(int)(a)) * (int)(b))
#  endif
#else
#  define mulunsign(a, b) (((ll)(ucell)(a)) * (ucell)(b))
#  ifdef hppa
#    define mulsign(a, b) \
     ({ \
	dcell result; \
	result.d=mulunsign(a,b); \
	if(a<0) result.hl.hi-=(b); \
	if(b<0) result.hl.hi-=(a); \
	result.d; \
     })
#  else
#    define mulsign(a, b) (((ll)(int)(a)) * (int)(b))
#  endif
#endif

static ucell revadd(ucell a, ucell b)
{
   int t;
   while(t=a&b, b=a^b, a=t) a>>=1;
   return b;
}

#define GETNOS(alu,sp,tos,nos,rol,fl,fr,sr,ml,mr) \
{ \
    if(ALUTYP(alu)) nos=(IMVAL(alu)); \
    else if(alu < 0x1A0) { \
      switch(ALUMOD(alu)) { \
	  case  0: nos=0; break; \
	  case  1: nos=-1; break; \
	  case  2: nos= 0x7FFFFFFF; break; \
	  case  3: nos=-0x80000000; break; \
	  case  4: nos=c0; break; \
	  case  5: nos=c1; break; \
	  case  6: nos=c2; break; \
	  case  7: nos=c3; break; \
	    \
	  default: \
	  case  8: nos=tos; rol=0; break; \
	  case  9: nos=SPT(sp+1); rol=1; break; \
	  case 10: nos=SPT(sp+2); rol=2; break; \
	  case 11: nos=SPT(sp+3); rol=3; break; \
	    \
	  case 12: nos=SPT(sp+4); break; \
	  case 13: nos=SPT(sp+5); break; \
	  case 14: nos=SPT(sp+6); break; \
	  case 15: nos=SPT(sp+7); break; \
	    \
	  case 16: nos=tos0; break; \
	  case 17: nos=SPT(sp0+1); break; \
	  case 18: nos=SPT(sp0+2); break; \
	  case 19: nos=SPT(sp0+3); break; \
	    \
	  case 20: nos=tos1; break; \
	  case 21: nos=SPT(sp1+1); break; \
	  case 22: nos=SPT(sp1+2); break; \
	  case 23: nos=SPT(sp1+3); break; \
	    \
	  case 24: nos=tos2; break; \
	  case 25: nos=SPT(sp2+1); break; \
	  case 26: nos=SPT(sp2+2); break; \
	  case 27: nos=SPT(sp2+3); break; \
	    \
	  case 28: nos=tos3; break; \
	  case 29: nos=SPT(sp3+1); break; \
	  case 30: nos=SPT(sp3+2); break; \
	  case 31: nos=SPT(sp3+3); break; \
	} \
    } else switch(alu) { \
	case 0x1D0:  nos=((cell)sp)-(cell)image; break; \
	case 0x1D1:  nos=(cell)loops; break; \
	case 0x1D2:  nos=(cell)loope; break; \
	case 0x1D3:  nos=(cell)ip; break; \
	case 0x1D4:  nos=sr|ssr; break; \
	case 0x1D5:  nos=(sr >> 8) | ((sr << 12) & 0x700); break; \
	case 0x1D6:  nos=index; break; \
	case 0x1D7:  fr.df=fl.df; break; \
	case 0x1EA: \
	case 0x1E4: \
	case 0x1E5:  fra.df=fl0.df+fl1.df; break; \
	case 0x1EB: \
	case 0x1E6: \
	case 0x1E7:  frm.df=fl2.df*fl3.df; break; \
	case 0x1A0: \
	case 0x1A1: \
	case 0x1A2: \
	case 0x1A3: \
	case 0x1A4: \
	case 0x1A5: \
	case 0x1A6: \
	case 0x1A7: \
	case 0x1A8: \
	case 0x1A9: \
	case 0x1AA: \
	case 0x1AB: \
	case 0x1AC: \
	case 0x1AD: \
	case 0x1AE: \
	case 0x1AF:  mr.d=ml.d; \
	default: nos=SPT(sp+1); \
     } \
}

#define ROL(rol,sp,tos) \
switch(rol) { \
   case 3: SPT(sp+3) = SPT(sp+2); \
   case 2: SPT(sp+2) = SPT(sp+1); \
   case 1: \
   case 0: rol=-1; sp++; \
}

#define GETFLAG(reg,tos,mode,sr) \
   switch(mode & 7) { \
	default: \
	case 0:  reg = -1; break; \
	case 1:  reg = -(tos == 0); break; \
	case 2:  reg = -(tos <  0); break; \
	case 3:  reg = -((sr >> 1) & 1); break; \
	case 4:  reg = (sr & 1) - 1; break; \
	case 5:  reg = -((sr & 1) & (tos != 0)); break; \
	case 6:  reg = -((tos < 0) ^ ((sr >> 1) & 1));  break; \
	case 7:  reg = -((tos != 0) & ((tos < 0) == ((sr >> 1) & 1))); \
	break; \
     } \
   if(mode & 8) reg = ~reg; \

#ifdef WITH_DIVISION
#  define DIVISION_MAKRO(sr,sp,tos,nos) \
	       case 8:  sr = ((sr & ~0x30) | \
			      (((tos < 0) * 0x20) ^ \
			       ((nos < 0) * 0x30))); \
	                if(tos < 0)  tos = -tos; \
	                if(nos < 0) \
		          { \
			     dcell tmp; \
			     tmp.hl.hi = nos; tmp.hl.lo = SPT(sp+2); \
			     tmp.d = -tmp.d; \
			     SPT(sp+1) = tmp.hl.hi; \
			     SPT(sp+2) = tmp.hl.lo; \
			  } \
	                break; \
	       case 9:  { \
			   int i, tail=SPT(sp+2); \
			   \
			   for(i=0; i<4; i++) \
			     { \
				if(nos < 0) \
				  { \
				     nos = (nos << 1) | (((ucell)tail) >> 31); \
				     tail += tail+1; \
				     nos -= tos; \
				  } \
				else \
				  { \
				     nos = (nos << 1) | (((ucell)tail) >> 31); \
				     tail += tail; \
				     if((ucell) tos < (ucell) nos) \
				       { \
					  nos -= tos; \
					  tail++; \
				       } \
				  } \
			     } \
			   SPT(sp+2) = tail; \
			   SPT(sp+1) = nos; \
			} break; \
	      case 10:  sp++; tos = (sr & 0x10) ? -nos : nos; \
	                SPT(sp+1) = (sr & 0x20) ? -SPT(sp+1) : SPT(sp+1); \
	                sr &= ~0x30; break; \
	      case 11:  sp++; \
	                if(sr & 0x20) \
		          { \
			     if(nos) \
			       { \
				  SPT(sp+1)=-(SPT(sp+1)+1); \
				  nos = tos - nos; \
			       } \
			     else \
			       SPT(sp+1) = -SPT(sp+1); \
			  } \
	                if(sr & 0x10) \
			  tos = -nos; \
	                else \
	                  tos = nos; \
                        sr &= ~0x30; \
	                break;
#else
#  define DIVISION_MAKRO(sr,sp,tos,nos) \
	      case  8:  /* ff1 */ \
	                SPT(sp--)=tos; \
		 { \
		     int i=0; \
		     if(tos==0) i=32; \
		     else while(((int)tos)>=0) i++, tos+=tos; \
		     tos=i; \
		 } break; \
	      case  9:  /* popc */ \
	                SPT(sp--)=tos; \
		 { \
		    int i=0; \
		    while(tos) i++, tos &= tos-1; \
		    tos=i; \
		 } break; \
	      case 10: ((ucell)tos) >>= 24; break; \
	      case 11: ((ucell)tos) >>= 16; break;
#endif

ucell rotl(ucell value, ucell count)
{
  count &= 0x1F;
  return ((value << count) |
	  (value >> (0x20 - count)));
}

#define EXEC(alu,tos,nos,sp,mlatch,mres,sr,fl,fal,fml) \
{ \
    if(ALUTYP(alu)) { \
	if(IMOP(alu)) tos= (tos << 8) + nos; \
	else { \
	  SPT(sp--)=tos; \
	  tos=(int)(char)nos; \
	} \
    } \
    else switch(ALUOP(alu)) { \
      case  0:  tos = tos | nos; break; \
      case  2:  if(sr & 1) { \
	    ll _a = (ll)tos + nos + 1; \
	    sr = ((((ucell) tos) > ((ucell) -nos)) | \
		  ((HI(_a) != -(((int)_a) < 0)) << 1) | \
		  (sr & -4)); \
	    tos = (cell)_a;  break; \
	 } \
      case  1:  sr = ((((ucell) tos) >= ((ucell) (-nos))) | \
		      ((tos <= ~nos) << 1) | \
		      (sr & -4)); \
	tos = tos + nos; \
	sr ^= (tos < 0) << 1; break; \
      case  3:  mlatch.d = mulsign(tos,nos); \
	++sp;  tos = SPT(sp); break; \
      case  4:  tos = tos & nos; break; \
      case  6: if(sr & 1) { \
	    sr = ((((ucell) tos) > ((ucell) nos)) | \
		  ((tos <= nos) << 1) | \
		  (sr & -4)); \
	    tos = tos - nos + 1; \
	    sr ^= (tos < 0) << 1;  break; \
	 } \
      case  5:  sr = ((((ucell) tos) >= ((ucell) nos)) | \
		      ((tos < nos) << 1) | \
		      (sr & -4)); \
	tos = tos - nos; \
	sr ^= (tos < 0) << 1; break; \
      case  7:  mlatch.d = mulunsign(tos,nos); \
	++sp;  tos = SPT(sp); break; \
      case  8:  tos = tos ^ nos; break; \
      case 10: if(sr & 1) { \
	    sr = ((((ucell) nos) > ((ucell) tos)) | \
		  ((nos <= tos) << 1) | \
		  (sr & -4)); \
	    tos = nos - tos + 1; \
	    sr ^= (tos < 0) << 1;  break; \
	 } \
      case  9:  sr = ((((ucell) nos) >= ((ucell) tos)) | \
		      ((nos < tos) << 1) | \
		      (sr & -4)); \
	tos = nos - tos; \
	sr ^= (tos < 0) << 1; break; \
      case 11:  mlatch.hl.hi=tos; mlatch.hl.lo=nos; \
	tos = SPT(++sp); break; \
      case 12:  /* pick, pin */ \
	if(ALUMOD(alu) < 8) { \
	    SPT(sp+ALUMOD(alu))=tos; \
	    tos=SPT(++sp); \
	} else { \
	    if(ALUMOD(alu) != 8) SPT(sp)=tos; sp--; \
	    tos=nos; \
	} \
	break; \
      case 13:  /* mul@, flag */ \
	{ \
	    int _alu=ALUMOD(alu); \
	    if(_alu < 16) { \
		ll _mres = (_alu & 4) ? -mres.d : mres.d; \
		int _mode  = (sr >> 4) & 7; \
		int _count = (int)((signed char)(sr >> 8)); \
		sr &= -4; \
		if(_alu & 1)  _mres = ((_count < 0) ? \
				      (_mode & 4 ? \
				       _mres >> -_count : \
				       ((ull)_mres) >> -_count) : \
				      _mres << _count); \
		if(_alu & 8) { \
		    dcell _d; _d.hl.hi=tos; _d.hl.lo=nos; _mres += _d.d; \
		} else { \
		  SPT(sp)=tos; sp -= 2; } \
		tos = (cell) HI(_mres); \
		if(_alu & 2) { \
		    switch(_mode) { \
		      case 4: \
		      case 0: if(LO(_mres) <= 0x80000000) \
			        tos += (tos & (LO(_mres)>>31)); \
		              else tos++; break; \
		      case 5: \
		      case 1: if(tos >= 0) break; \
		      case 6: \
		      case 2: if(LO(_mres)) tos++; \
		      default: break;\
		    } \
		    sp++; \
		} else { \
		    SPT(sp+1)=(cell) LO(_mres); \
		} \
	    } else { \
	       GETFLAG(tos,tos,_alu,sr) \
	    } \
	} break; \
      case 14:  /* shift, div, regs */ \
	switch(ALUMOD(alu)) { \
	  case 0:  sr = (tos & 1) | (sr & -4); \
	    tos=((int)  tos) >> 1; break; \
	  case 1:  sr = (tos & 1) | (sr & -4); \
	    tos=((ucell) tos) >> 1; break; \
	  case 2:  sr = (tos & 1) | (sr & -4); \
	    tos=(((ucell) tos) >> 1) | (tos << 31); \
	    break; \
	  case 3:  { \
		       int _sr=sr & 1; \
		       sr = (tos & 1) | (sr & -4); \
		       tos= ((((ucell) tos) >> 1) | \
			     (_sr << 31)); \
		   } break; \
	  case 4:  sr = (tos < 0) | (sr & -4); \
	    tos=tos << 1; break; \
	  case 5:  sr = (tos < 0) | (sr & -4); \
	    tos=tos << 1; \
	    sr |= ((sr & 1) ^ (tos < 0)) << 1; \
	    break; \
	  case 6:  sr = (tos < 0) | (sr & -4); \
	    tos=(tos << 1) | ((ucell)tos >> 31); break; \
	  case 7:  { \
		       int _sr=sr & 1; \
		       sr = (tos < 0) | (sr & -4); \
		       tos= (tos << 1) | _sr; \
		   } break; \
	    \
	    DIVISION_MAKRO(sr,sp,tos,nos) \
	    \
	  case 12:  tos |= -(tos & 0x80); break; \
	  case 13:  tos |= -(tos & 0x8000); break; \
	  case 14:  tos <<= 24; break; \
	  case 15:  tos <<= 16; break; \
	    \
	  case 16: \
	  case 17: \
	  case 18: \
	  case 19: \
	  case 20: \
	  case 21: \
	  case 22:  SPT(sp--)=tos; tos=nos; break; \
	  case 23:  SPT(sp--)=tos; \
	    SPT(sp--)=fl.hl.lo; \
	    tos    =fl.hl.hi; break; \
	  case 24:  /* ssr|=EXTDEBUG;*/ \
		    sp=(cell *)(tos-sizeof(cell)+(cell)image); \
		    tos=SPT(++sp); break; \
	  case 25:  loops=(opcode *)(tos & -8);   tos=SPT(++sp); break; \
	  case 26:  loope=(opcode *)(tos & -8);   tos=SPT(++sp); break; \
	  case 27:  ip   =(opcode *)(tos)-1; tos=SPT(++sp); break; \
	  case 28:  sr=tos & 0x0FFFF; \
	    if(ssr & SUPERVISOR) ssr=tos & 0x0FFFF0000; \
	    tos=SPT(++sp); break; \
	  case 29:  sr=(((tos & 0x0FF) << 8) | \
			(((tos<0 ? tos ^ 0x300 : tos) >> 4) & 0x70) | \
			(sr & 0x0FFFF008F)); \
	    tos=SPT(++sp); break; \
	  case 30:  index=tos;  tos=SPT(++sp); break; \
	  case 31:  fl.hl.hi=tos;  fl.hl.lo=nos; sp+=2; tos=SPT(sp); \
	    break; \
	} \
	break; \
      case 15:  /* FP, bit field */ \
	switch(ALUMOD(alu)) { \
	  case  1:  tos ^= -0x80000000; \
	  case  0:  fal.hl.hi = tos; fal.hl.lo = nos; \
	    sp+=2; tos=SPT(sp); break; \
	  case  3:  tos ^= -0x80000000; \
	  case  2:  fml.hl.hi = tos; fml.hl.lo = nos; \
	    sp+=2; tos=SPT(sp); break; \
	  case  4:  fal.df= fra.df; break; \
	  case  5:  fal.df=-fra.df; break; \
	  case  6:  fal.df= frm.df; break; \
	  case  7:  fal.df=-frm.df; break; \
	  case  8:  fal.hl.hi = 0x43300000; \
	    fal.hl.lo = -0x80000000 ^ tos; \
	    tos=SPT(++sp); break; \
	  case  9:  fal.hl.hi = 0xC3300000; \
	    fal.hl.lo = -0x80000000 ^ tos; \
	    tos=SPT(++sp); break; \
	  case 10:  SPT(sp--)=tos; \
	    SPT(sp--)=fra.hl.lo; \
	    tos    =fra.hl.hi; break; \
	  case 11:  SPT(sp--)=tos; \
	    SPT(sp--)=frm.hl.lo; \
	    tos    =frm.hl.hi; break; \
	  case 12:  { \
			union { cell ss; float sf; } tmp; \
			dcell _d; \
			tmp.ss = tos; \
			_d.df  = (double) tmp.sf; \
			SPT(sp--)= _d.hl.lo; \
			tos    = _d.hl.hi; \
		    } break; \
	  case 13:  { \
			union { cell ss; float sf; } tmp; \
			dcell _d; \
			_d.hl.hi=tos; _d.hl.lo=nos; \
			tmp.sf = (float)_d.df; \
			tos=tmp.ss; sp++; \
		    } break; \
	  case 14:  SPT(sp--)= (tos & 0x800FFFFF) | 0x3FF00000; \
	    tos= ((tos >> 20) & 0x7FF) - 0x3FF; break; \
	    /* bit operations, */ \
	    /* tos in format: 0, rot mask <<, mask len <<, rot source <<*/\
	  case 15:  tos=nos+(tos << 20); sp++; break; \
	  case 16:  { \
			int mask = -1; \
			nos = rotl(nos,tos); \
			tos >>= 8; \
			if(tos) { \
			  mask = ~(mask << (tos & 0x1F)); \
			  tos >>= 8; \
			  mask = rotl(mask,tos); \
			  tos = mask & nos; \
			} else tos=nos; \
		    } sp++; break; \
	  case 17:  { \
			unsigned int mask = -1; \
			nos = rotl(nos,tos); \
			tos >>= 8; \
			if(tos) { \
			  mask = ~(mask << (tos & 0x1F)); \
			  tos >>= 8; \
			  mask = rotl(mask,tos); \
			  tos = mask & nos; \
			  mask = ((~mask)+1) & (~mask); \
			  if ((mask >> 1) & tos)  tos |= mask; \
			} else tos=nos; \
		    } sp++; break; \
	  case 18:  break; /* padd */ \
	  case 19:  break; /* psub */ \
	  case 20:  break; /* cc@ */ \
	  case 21:  break; /* cc! */ \
	  case 23:  break; /* pdiff@ */ \
	  case 22:  break; /* pdiff! */ \
	  case 24:  /* pixel expand 1 bit */ \
	    { \
		int i; dcell r; \
		for(i=0; i<32; i++) { \
		    r.d <<= 2; \
		    r.hl.lo |= (((tos >> 30) & 0x2) | \
				((nos >> 31) & 0x1)); \
		    tos <<= 1; nos <<= 1; \
		} \
		tos=r.hl.hi; \
		SPT(sp+1)=r.hl.lo; \
	    } break; \
	  case 25:  /* pixel expand 2 bit */ \
	    { \
		int i; dcell r; \
		for(i=0; i<16; i++) { \
		    r.d <<= 4; \
		    r.hl.lo |= (((tos >> 28) & 0xC) | \
				((nos >> 30) & 0x3)); \
		    tos <<= 2; nos <<= 2; \
		} \
		tos=r.hl.hi; \
		SPT(sp+1)=r.hl.lo; \
	    } break; \
	  case 26:  /* pixel expand 4 bit */ \
	    { \
		int i; dcell r; \
		for(i=0; i<8; i++) { \
		    r.d <<= 8; \
		    r.hl.lo |= (((tos >> 24) & 0xf0) | \
				((nos >> 28) & 0x0f)); \
		    tos <<= 4; nos <<= 4; \
		} \
		tos=r.hl.hi; \
		SPT(sp+1)=r.hl.lo; \
	    } break; \
	  case 27:  /* pixel expand 8 bit */ \
	    { \
		int i; dcell r; \
		for(i=0; i<4; i++) { \
		    r.d <<= 16; \
		    r.hl.lo |= (((tos >> 16) & 0xff00) | \
				((nos >> 24) & 0x00ff)); \
		    tos <<= 8; nos <<= 8; \
		} \
		tos=r.hl.hi; \
		SPT(sp+1)=r.hl.lo; \
	    } break; \
	  case 28:  /* pixel pack 1 bit */ \
	    { \
		int i; dcell r; \
		r.hl.hi=tos; r.hl.lo=nos; \
		for(i=0; i<32; i++) { \
		    tos <<= 1; nos <<= 1; \
		    tos |= (r.hl.hi >> 31) & 0x1; \
		    nos |= (r.hl.hi >> 30) & 0x1; \
		    r.d <<= 2; \
		} \
		SPT(sp+1)=nos; \
	    } break; \
	  case 29:  /* pixel pack 2 bit */ \
	    { \
		int i; dcell r; \
		r.hl.hi=tos; r.hl.lo=nos; \
		for(i=0; i<16; i++) { \
		    tos <<= 2; nos <<= 2; \
		    tos |= (r.hl.hi >> 30) & 0x3; \
		    nos |= (r.hl.hi >> 28) & 0x3; \
		    r.d <<= 4; \
		} \
		SPT(sp+1)=nos; \
	    } break; \
	  case 30:  /* pixel pack 4 bit */ \
	    { \
		int i; dcell r; \
		r.hl.hi=tos; r.hl.lo=nos; \
		for(i=0; i<8; i++) { \
		    tos <<= 4; nos <<= 4; \
		    tos |= (r.hl.hi >> 28) & 0xF; \
		    nos |= (r.hl.hi >> 24) & 0xF; \
		    r.d <<= 8; \
		} \
		SPT(sp+1)=nos; \
	    } break; \
	  case 31:	/* pixel pack 8 bit */ \
	    { \
		int i; dcell r; \
		r.hl.hi=tos; r.hl.lo=nos; \
		for(i=0; i<4; i++) { \
		    tos <<= 8; nos <<= 8; \
		    tos |= (r.hl.hi >> 24) & 0xFF; \
		    nos |= (r.hl.hi >> 16) & 0xFF; \
		    r.d <<= 16; \
		} \
		SPT(sp+1)=nos; \
	    } break; \
	} \
	break; \
    } \
}

#if 0
 /* tos in the format length_6,scale_2,pos_6 */ \
 case 16:  nos >>= (32-(tos & 0x1F)-((tos >> 8) & 0x1F)); \
           tos >>= 6; \
           if(tos >> 2) { \
	     int tmp = tos >> 2; \
             nos &= (1 << tmp) - 1 ; \
             nos |= -(nos & (1 << (tmp-1))); \
           } \
           tos = nos << (tos & 3);  sp++; break; \
 case 17:  nos = (ucell)nos >> (32-(tos & 0x1F)-((tos >> 8) & 0x1F)); \
           tos >>= 6; \
           if(tos >> 2) \
           { \
	       int tmp = tos >> 2; \
               nos &= (1 << tmp) - 1 ; \
           } \
           tos = nos << (tos & 3);  sp++; break; \
 case 18:  { \
	       int tmp = (tos >> 8) & 0x1F; \
	       tos &= 0x1F; \
	       if(tmp) \
	       nos &= (1 << tmp) - 1 ; \
	       nos <<= tos; \
	       tos = nos; sp++; \
	   } break; \
 case 19:  { \
	       int tmp = (tos >> 8) & 0x1F; \
	       int mask = -1; \
	       tos &= 0x1F; \
	       if(tmp) mask &= (1 << tmp) - 1 ; \
	       mask <<= tos; mask |= (1 << tos) - 1; \
	       tos = nos & mask; sp++; \
	   } break;
#endif

#define PUSHLD(val,num,s0l,s0h,spl,sph,srl,srh) \
switch(num) \
{ \
  case 3:  if(EXECQ(srh)) { SPT(sph--)=s0h; s0h=val[1]; } \
  case 1:  if(EXECQ(srl)) { SPT(spl--)=s0l; s0l=val[0]; } \
    num=0; break; \
  case 4:  if(EXECQ(srh)) \
    { SPT(sph--)=s0h; SPT(sph--)=val[3]; s0h=val[2]; } \
  case 2:  if(EXECQ(srl)) \
    { SPT(spl--)=s0l; SPT(spl--)=val[1]; s0l=val[0]; } \
    num=0; break; \
  case 7:  if(EXECQ(srl)) { SPT(spl--)=s0l; s0l=val[1]; } \
  case 5:  if(EXECQ(srh)) { SPT(sph--)=s0h; s0h=val[0]; } \
    num=0; break; \
  case 8:  if(EXECQ(srl)) \
    { SPT(spl--)=s0l; SPT(spl--)=val[3]; s0l=val[2]; } \
  case 6:  if(EXECQ(srh)) \
    { SPT(sph--)=s0h; SPT(sph--)=val[1]; s0h=val[0]; } \
    num=0; break; \
  case 0:  break; \
}

#define UPDAT(dat,tos,reg,imm,val,num,spl,sph,s0l,s0h,srl,srh) \
{ \
   cell n=(dat & 3), mode=((dat >> 2) & 7), \
        c=((dat >> 5) & 7), sh=STACKSIDE(dat); \
   memreg *rm=reg+(mode & 3), *rn=reg+n; \
   cell m=rn->m, f=rn->f, r=rn->r, mask=(-1 << (f >> 8)), rs; \
   \
   switch(c) \
     { \
	case 0: if(f & 4) { \
		      ucell a=(mode & 4) ? tos : (sh ? rn->n : 0); \
		      rs = revadd(revadd(r,a),imm); \
		   } else { \
                      rs = r + ((mode & 4) ? tos : (sh ? rn->n : 0)) + imm; \
	              if(m && rs >= m) rs = r & mask; \
		   } \
	        rm->r = rs; break; \
	case 1: if(f & 4) { \
	              ucell a=((mode & 4) ? tos : (sh ? rn->n : 0)); \
		      rs = revadd(revadd(r,revadd(~a,0x80000000)),imm);\
		   } else { \
   	              rs = r - ((mode & 4) ? tos : (sh ? rn->n : 0)) + imm; \
	              if(m && rs >= m) rs = r & mask; \
		   } \
	        rm->r = rs; break; \
        case 2: \
	case 3: if(sh ? EXECQ(srh) : EXECQ(srl)) { \
		   cell ntos; /* memreg rn */\
		   if(mode & 4) { \
		     if(c & 1) \
		       ntos = sh ? SPT(++sph) : SPT(++spl); \
		     else \
		       ntos = sh ? s0h : s0l; \
		     ((ucell*)rn)[mode & 3] = ntos; \
	             if(sh) s0h=SPT(++sph); \
	             else   s0l=SPT(++spl); \
		   } else { \
		     if(sh) (SPT(sph--) = s0h); \
		     else   (SPT(spl--) = s0l); \
		     ntos = ((ucell *)rn)[mode & 3]; \
		     if(c & 1) { \
		       (sh ? SPT(sph--) : SPT(spl--)) = ntos; \
		       (sh ? s0h : s0l) = 0; \
		     } else \
		       (sh ? s0h : s0l) = ntos; \
		   } \
		} \
	        break; \
	case 4: \
	case 5: break; \
	case 6: /* IO-Control */ \
	        if(sh ? EXECQ(srh) : EXECQ(srl)) { \
		   dcell io_data; \
		   switch(mode) { \
			case 0: /* inb  */ \
			        if(sh) SPT(sph--) = s0h; \
			        else   SPT(spl--) = s0l; \
				(sh ? s0h : s0l) = 0xff & ioin(rn->r + imm, image); \
			        break; \
			case 1: /* inh  */ \
			        if(sh) SPT(sph--) = s0h; \
			        else   SPT(spl--) = s0l; \
				(sh ? s0h : s0l) = 0xffff & ioin(rn->r + imm, image); \
			        break; \
			case 2: /* in   */ \
			        if(sh) SPT(sph--) = s0h; \
			        else   SPT(spl--) = s0l; \
				(sh ? s0h : s0l) = 0xffffffff & ioin(rn->r + imm, image); \
			        break; \
			case 3: /* ind  */ \
			        if(sh) SPT(sph--) = s0h; \
			        else   SPT(spl--) = s0l; \
				{ ull io_q = ioin(rn->r + imm, image); \
				    if(sh) SPT(sph--) = TLO(io_q); \
				    else   SPT(spl--) = TLO(io_q); \
				    (sh ? s0h : s0l) = THI(io_q); \
				    break; } \
			case 4: /* outb  */ \
			case 5: /* outh  */ \
			case 6: /* out   */ \
			        io_data.hl.hi=0; \
			        io_data.hl.lo=sh ? s0h : s0l; \
			        ioout(rn->r + imm, io_data.d, image); \
				if(sh) s0h=SPT(++sph); \
				else   s0l=SPT(++spl); break; \
			case 7: /* outd */ \
			        io_data.hl.hi=sh ? s0h : s0l; \
			        io_data.hl.lo=sh ? SPT(++sph) : SPT(++spl); \
			        ioout(rn->r + imm, io_data.d, image); \
			        if(sh) s0h=SPT(++sph); \
			        else   s0l=SPT(++spl); break; \
		     } \
		} break; \
	case 7: /* set/get special registers */ \
	        if(sh ? EXECQ(srh) : EXECQ(srl)) { \
		   dcell dvalue; \
		   \
		   switch(mode) { \
			case 0: \
			 switch(n) { \
			   case 0: c0=tos; break; \
			   case 1: c1=tos; break; \
			   case 2: c2=tos; break; \
			   case 3: c3=tos; break; \
			 } break; \
			case 1: \
			 switch(n) { \
			   default: break; \
			   case  0: dvalue.d=cpustate[imm]; \
			     if(sh) { \
				 *sph-- = s0h; \
				 *sph-- = dvalue.hl.lo; \
				 s0h    = dvalue.hl.hi; \
			     } else { \
				 *spl-- = s0l; \
				 *spl-- = dvalue.hl.lo; \
				 s0l    = dvalue.hl.hi; \
			     } break; \
			 } \
			case 7: switch(n) { \
			   case  0: if(sh) { \
				 dvalue.hl.hi = s0h; \
				 dvalue.hl.lo = SPT(++sph); \
				 s0h          = SPT(++sph); \
			     } else { \
				 dvalue.hl.hi = s0l; \
				 dvalue.hl.lo = SPT(++spl); \
				 s0l          = SPT(++spl); \
			     } \
			     cpustate[imm]=dvalue.d; \
			 } \
		     } \
		} break; \
     } \
}

#define GETIM(dat,dat2,imm) \
{ \
   if(dat & 0x100) imm=(int)((unsigned char) dat) + (imm << 8); \
   else            imm=(int)((signed char) dat); \
}

#define GETEA(dat,tos,reg,ea,imm,spl,sph,s0l,s0h,srl,srh,val,num,push) \
{ \
   cell n=(dat & 3), mode=((dat >> 2) & 7), s=((dat >> 5) & 3); \
   memreg *rn=reg+n; \
   cell i=imm, f=rn->f, r=rn->r, m=rn->m, mask=-(!!(f >> 8)) << (f >> 8), rs; \
   cell disp=((mode & 4) ? tos : rn->n); \
   \
   if(dat & 0x80) s++; \
   if((dat & 0x200) && (f & 0x10)) \
      if(f & 4)  disp = revadd(~disp,0x80000000); \
      else       disp = -disp; \
   if(f & 1) disp <<= s; \
   if(f & 4) rs = revadd(r, disp); \
   else      rs = r + disp; \
   if(m && rs >= m) rs = 0; \
   \
   switch(mode) { \
	case 0: ea = (void *) rn->r; break; \
	case 5: \
	case 1:	ea = (void *) rs; break; \
	case 6: \
	case 2: ea = (void *) rn->r; rn->r = rs; break; \
	case 7: \
	case 3: ea = (void *) rn->r = rs; break; \
	case 4: f = 1; mask = 0; \
	        switch(n) { \
			    case 0: ea = (void *) ip; break; \
			    case 2: ea = ((void *) ip) + disp; break; \
			    case 1: ea = (void *) tos; break; \
			    case 3: ea = (void *) tos; f |= 2; break; \
		} \
		break; \
     } \
   if(f & 1) i <<= s; \
   if(f & 4) { \
	ea = (void *) revadd((int) ea, i); \
     } \
   else ea += i; \
   \
   ea = (void *)(((cell)ea & ~mask) | (r & mask)); \
   rs = ((rs & ~mask) | (r & mask)); \
   \
   ea = (void *) ((long)(ea) & (-1 << s)); \
   \
   if(SWITCHEND(f)) ea = (void *) (((int) ea) ^ ((8 - (1 << s)) & 7)); \
   \
   if(dat & 0x200) { \
	if(dat & 0x80) { \
	     if(STACKSIDE(dat)) { \
		  switch(s) { \
		       case 1: if(EXECQ(srh)) PTR(((char *)ea)+HIH)=(char)s0h; \
		               if(EXECQ(srl)) PTR(((char *)ea)+LOH)=(char)s0l; break; \
		       case 2: if(EXECQ(srh)) PTR(((short *)ea)+HIH)=(short)s0h; \
		               if(EXECQ(srl)) PTR(((short *)ea)+LOH)=(short)s0l; break; \
		       case 3: if(EXECQ(srh)) PTR(((cell *)ea)+HIH)=s0h; \
		               if(EXECQ(srl)) PTR(((cell *)ea)+LOH)=s0l; break; \
		       case 4: if(EXECQ(srh)) {PTR(((cell *)ea)+HIH)=s0h; \
		                  PTR(((cell *)ea)+LOH)=SPT(++sph);} \
		               if(EXECQ(srl)) {PTR(((cell *)ea)+2+HIH)=s0l; \
		                  PTR(((cell *)ea)+2+LOH)=SPT(++spl);} break; \
		    } \
		  if(EXECQ(srh)) s0h=SPT(++sph); \
		  if(EXECQ(srl)) s0l=SPT(++spl); \
	       } else { \
		  switch(s) { \
		       case 1: if(EXECQ(srl)) PTR(((char *)ea)+HIH)=(char)s0l; \
		               if(EXECQ(srh)) PTR(((char *)ea)+LOH)=(char)s0h; break; \
		       case 2: if(EXECQ(srl)) PTR(((short *)ea)+HIH)=(short)s0l; \
		               if(EXECQ(srh)) PTR(((short *)ea)+LOH)=(short)s0h; break; \
		       case 3: if(EXECQ(srl)) PTR(((cell *)ea)+HIH)=s0l; \
		               if(EXECQ(srh)) PTR(((cell *)ea)+LOH)=s0h; break; \
		       case 4: if(EXECQ(srl)) {PTR(((cell *)ea)+HIH)=s0l; \
		                 PTR(((cell *)ea)+LOH)=SPT(++spl);} \
		               if(EXECQ(srh)) {PTR(((cell *)ea)+2+HIH)=s0h; \
		                 PTR(((cell *)ea)+2+LOH)=SPT(++sph);} break; \
		    } \
		  if(EXECQ(srh)) s0h=SPT(++sph); \
		  if(EXECQ(srl)) s0l=SPT(++spl); \
	       } \
	  } else { \
	     if(STACKSIDE(dat)) { \
		  switch(s) { \
		       case 0: if(EXECQ(srh)) PTR(((char *)ea))=(char)s0h; break; \
		       case 1: if(EXECQ(srh)) PTR(((short *)ea))=(short)s0h; break; \
		       case 2: if(EXECQ(srh)) PTR(((cell *)ea))=s0h; break; \
		       case 3: if(EXECQ(srh)) (PTR(((cell *)ea)+HIH)=s0h), \
		               PTR(((cell *)ea)+LOH)=SPT(++sph); break; \
		    } \
		  if(EXECQ(srh)) s0h=SPT(++sph); \
	       } else { \
		  switch(s) { \
		       case 0: if(EXECQ(srl)) PTR(((char *)ea))=(char)s0l; break; \
		       case 1: if(EXECQ(srl)) PTR(((short *)ea))=(short)s0l; break; \
		       case 2: if(EXECQ(srl)) PTR(((cell *)ea))=s0l; break; \
		       case 3: if(EXECQ(srl)) (PTR(((cell *)ea)+HIH)=(char)s0l), \
		               PTR(((cell *)ea)+LOH)=SPT(++spl); break; \
		    } \
		  if(EXECQ(srl)) s0l=SPT(++spl); \
	       } \
	  } \
	goto push; \
     } else { \
	PUSHLD(val,num,s0l,s0h,spl,sph,srl,srh); \
	\
	if(dat & 0x80) { \
	     num=3; \
	     switch(s) { \
		  case 1:  val[0]=(cell)PTR(((unsigned char *)ea)+HIH); \
		           val[1]=(cell)PTR(((unsigned char *)ea)+LOH); break; \
		  case 2:  val[0]=(cell)PTR(((unsigned short *)ea)+HIH); \
		           val[1]=(cell)PTR(((unsigned short *)ea)+LOH); break; \
		  case 3:  val[0]=PTR(((cell *)ea)+HIH); \
		           val[1]=PTR(((cell *)ea)+LOH); break; \
		  case 4:  num++; \
		           val[0]=PTR(((cell *)ea)+HIH); \
		           val[1]=PTR(((cell *)ea)+LOH); \
		           val[2]=PTR(((cell *)ea)+2+HIH); \
		           val[3]=PTR(((cell *)ea)+2+LOH); break; \
	       } \
	  } else { \
	     num=1; \
	     switch(s) { \
		  case 0:  val[0]=(cell)PTR(((unsigned char *)ea)); break; \
		  case 1:  val[0]=(cell)PTR(((unsigned short *)ea)); break; \
		  case 2:  val[0]=PTR(((cell *)ea)); break; \
		  case 3:  num++; \
		           val[0]=PTR(((cell *)ea)+HIH); \
		           val[1]=PTR(((cell *)ea)+LOH); break; \
	       } \
	  } \
	if(STACKSIDE(dat)) num+=4; \
     } \
}

#define GETSP(dat,tos,s0l,s0h,srl,srh,imm) \
{ \
   if(STACKSIDE(dat)) { \
	tos=EXECQ(srh)? s0h : 0; \
     } else { \
	tos=EXECQ(srl)? s0l : 0; \
     } \
   imm=0; \
}

#define PSTACK(n,sp,tos) \
fprintf(stderr,"%ds: %8x %8x %8x %8x  %8x %8x %8x %8x\n",n,tos, \
       SPT(sp+1),SPT(sp+2),SPT(sp+3),SPT(sp+4),SPT(sp+5),SPT(sp+6),SPT(sp+7));

ll cpustate[256] = 
{
   1, 1, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,

   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,


   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,

   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
   0, 0, 0, 0,  0, 0, 0, 0,   0, 0, 0, 0,  0, 0, 0, 0,
};

#if defined(hppa) && defined(FORCE_REG)
#  define SP0REG  asm("%r6")
#  define SP1REG  asm("%r7")
#  define SP2REG  asm("%r8")
#  define SP3REG  asm("%r9")

#  define TOS0REG  asm("%r10")
#  define TOS1REG  asm("%r11")
#  define TOS2REG  asm("%r12")
#  define TOS3REG  asm("%r13")

#  define NOS0REG  asm("%r14")
#  define NOS1REG  asm("%r15")
#  define NOS2REG  asm("%r16")
#  define NOS3REG  asm("%r17")

#  define OPREG    asm("%r4")
#endif

#ifndef SP0REG
#  define SP0REG
#endif
#ifndef SP1REG
#  define SP1REG
#endif
#ifndef SP2REG
#  define SP2REG
#endif
#ifndef SP3REG
#  define SP3REG
#endif

#ifndef TOS0REG
#  define TOS0REG
#endif
#ifndef TOS1REG
#  define TOS1REG
#endif
#ifndef TOS2REG
#  define TOS2REG
#endif
#ifndef TOS3REG
#  define TOS3REG
#endif

#ifndef NOS0REG
#  define NOS0REG
#endif
#ifndef NOS1REG
#  define NOS1REG
#endif
#ifndef NOS2REG
#  define NOS2REG
#endif
#ifndef NOS3REG
#  define NOS3REG
#endif

#ifndef OPREG
#  define OPREG
#endif

#ifdef CATCH_INT
cell ssr=SUPERVISOR;

static void
debug_on (int sig)
{
  ssr |= EXTDEBUG;
  signal(sig,debug_on);
}
#endif

void emulate(ull *image, opcode *ip,
	     cell *vsp0, cell *vsp1, cell *vsp2, cell *vsp3)
{
  register opcode op OPREG, *nextip=ip, *brip=0;
  register cell *sp0 SP0REG =&PTR(vsp0), *sp1 SP1REG =&PTR(vsp1),
                *sp2 SP2REG =&PTR(vsp2), *sp3 SP3REG =&PTR(vsp3);
  register cell tos0 TOS0REG =SPT(sp0), tos1 TOS1REG =SPT(sp1),
                tos2 TOS2REG =SPT(sp2), tos3 TOS3REG =SPT(sp3);
  register cell nos0 NOS0REG =0, nos1 NOS1REG =0,
                nos2 NOS2REG =0, nos3 NOS3REG =0;
  dcell fl0, fl1, fl2, fl3;
  dcell fr0, fr1, fr2, fr3;
  dcell fra, frm;
  dcell ml0, ml1, mr0, mr1;
  register cell sr0=0, sr1=0, sr2=0, sr3=0;
  register cell tosr0=0, tosr1=0, tosr2=0, tosr3=0;
#ifndef CATCH_INT
  register cell ssr=SUPERVISOR;
#endif
  register opcode *loops=0, *loope=0;
  register cell index=0;
  register cell c0=1, c1=2, c2=4, c3=0x100000;
  register ucell op0=NOOP, op1=NOOP, op2=NOOP, op3=NOOP;
  register ucell dat0=NOOP, dat1=NOOP;
  register int  rol0=-1, rol1=-1, rol2=-1, rol3=-1;

/* Data move registers */
  memreg re[4] /*=(memreg *)malloc(sizeof(memreg)*4)*/;
  memreg ro[4] /*=(memreg *)malloc(sizeof(memreg)*4)*/;
  void *eaevn=NULL; cell imevn=0, numevn=0, valevn[4];
  void *eaodd=NULL; cell imodd=0, numodd=0, valodd[4];
  register cell s0e=tos0, s0o=tos1;

  fl0.df=0.; fl1.df=0.; fl2.df=0.; fl3.df=0.;
  fr0.df=0.; fr1.df=0.; fr2.df=0.; fr3.df=0.;
  fra.df=0.; frm.df=0.;
  ml0.d=0; ml1.d=0; mr0.d=0; mr1.d=0;

/* Initialize memory unit registers */

#ifdef CATCH_INT
  signal(SIGINT, debug_on);
#endif

  for(index=0; index<4; index++) {
    re[index].r=0; ro[index].r=0;
    re[index].n=1; ro[index].n=1;
    re[index].m=0; ro[index].m=0;
    re[index].f=1; ro[index].f=1;
  }
  index = 0;

/* Main loop: Simulate while IP != 0 */

  while(nextip != 0) {

/* CPU step state change */
    
    nextip = (opcode *)(-8 & (long)nextip);
    op=PTR(nextip++);
    cpustate[0]++;
    cpustate[1]++;
    cpustate[2]+=42;
    ip++;

/* redraw screen every w*h*2 instructions */

#ifdef WITH_SCREEN
    if(dpy && !(pixels--)) {
      check_event(2);
      pixels = w*h*4;
    }
#endif

/* check for external trace mode and debug external, if set */

    if(ssr & EXTDEBUG) {
      int i; static char c='s'; static char oldc='s';
      static Label debug=&&firstdebug;
      static opcode *skipto;
      static int nextarg=0, skips=1;
      
      goto *debug;
    firstdebug:
      fprintf(stderr,"\033[2J");
      debug=&&normaldebug;
      
    normaldebug:
      skips--;
      if(skips > 0) goto enddebug;
      skips=1;
      fprintf(stderr,"\033[%i;1H%8x: ",LINES,(int)(nextip-1));
      disass(op);
      
      fprintf(stderr,"\033[0;0Hloops: %8x, loope: %8x, index: %8x\n",
	     (int)loops,(int)loope,index);
      fprintf(stderr,"sr:    %8x, %8x, %8x, %8x\n",ssr|sr0,ssr|sr1,ssr|sr2,ssr|sr3);
      fprintf(stderr,"Stack: %8x, %8x, %8x, %8x\033[K\n",
	     (int)sp0-(int)image,(int)sp1-(int)image,
	     (int)sp2-(int)image,(int)sp3-(int)image);
      
      PSTACK(0,sp0,tos0);
      PSTACK(1,sp1,tos1);
      PSTACK(2,sp2,tos2);
      PSTACK(3,sp3,tos3);
      
      for(i=0; i<4; i++) {
	fprintf(stderr,"R%d: %8x %8x %8x %4x  R%d: %8x %8x %8x %4x\n",
	       i,re[i].r,re[i].n,re[i].m,re[i].f,
	       i,ro[i].r,ro[i].n,ro[i].m,ro[i].f);
      }
      /*
	fprintf(stderr,"EA: %8x (+%x)       | EAo: %8x (+%x)\n",
	(int)eaevn,imodd,(int)eaodd,imevn);
	fprintf(stderr,"e: %8x %8x %8x %8x | o: %8x %8x %8x %8x\n",
	valevn[0],valevn[1],valevn[2],valevn[3],
	valodd[0],valodd[1],valodd[2],valodd[3]);
	*/
      fprintf(stderr,"\033[M\033[%i;1H",LINES);
      fflush(stdout);

#ifdef WITH_SCREEN
      if(dpy) {
	int what=2;
	do {
	  check_event(what);
	  what = 1;
	}
#   ifndef MSDOS
	while(key_avail(stdin)==0);
#   else
        while(kbhit()==0);
#   endif
      }
      oldc=c;
#endif

    getkey:
      if(nextarg)
	fprintf(stderr,"\033[%d;1H%d",LINES,nextarg);
      else
	fprintf(stderr,"\033[%d;1H%c\010",LINES,c);
      fflush(stdout);
#   ifndef MSDOS
      c=getkey(stdin);
#   else
      c=getkey();
#   endif
    switchkey:
      switch(c) {
      case '?': nextarg=0;
#ifdef MSDOS
#define help(c,s) fprintf(stderr,"\033[12;1H\033[M\033[%i;1H\033[1m%c\033[0m %s\n",LINES,c,s)
#else
#define help(c,s) fprintf(stderr,"\033[12;1H\033[M\033[%i;1H\033[1m%c\033[0m %s",LINES,c,s)
#endif
	help('c',"Continue");
	help('e',"Continue till end of loop");
	help('q',"Quit");
	help('s',"Single Step");
	help('S',"Set Breakpoint at next addr and continue");
	help('?',"Help");
	goto getkey;
      case 'c':  nextarg=0; ssr &= ~EXTDEBUG; break;
      case 'q':  nextarg=0; fprintf(stderr,"Really quit? (y/n)");
	fflush(stdout);
	if(
#ifndef MSDOS
	   getkey(stdin)
#else
	   getkey()
#endif
	   =='y') return;
      case 's':  debug=&&normaldebug; skips=nextarg; nextarg=0; break;
      case 'S':  debug=&&skipnext; skipto=nextip+1;
	skips=nextarg; nextarg=0; break;
      case 'e':  debug=&&loopend; skips=nextarg; nextarg=0; break;
      case '0' ... '9': nextarg=nextarg*10+(c-'0'); goto getkey;
      case '\010':
      case '\177': nextarg=nextarg/10; goto getkey;
      default :  if(c>' ') {
	c=oldc;
	putchar(7);
	goto getkey;
      }
      c=oldc;
      oldc=0xff;
      goto switchkey;
      }
      goto enddebug;
  skipnext:
      if(skipto==nextip) {
	debug=&&normaldebug;
	goto normaldebug;
      }
      goto enddebug;
  loopend:
      if((ip > loope) || (ip < loops)) {
	debug=&&normaldebug;
	goto normaldebug;
      }
      goto enddebug;
  enddebug:  ;
  }

/* decode instruction */

  switch(TYP(op)) {
  case 0:
    dat1=DATA1(op);
    dat0=DATA0(op);
    if(CU0(op) || (dat0 & 0x200))
      GETSP(dat0,s0e,tos0,tos2,sr0,sr2,imevn)
      else  GETIM(dat0,dat1,imevn);
    
    if(CU1(op) || (dat1 & 0x200))
      GETSP(dat1,s0o,tos1,tos3,sr1,sr3,imodd)
      else  GETIM(dat1,dat0,imodd);
  case 1:
  getall:
  op3=(sr3 & 8) ? NOOP : ALU3(op);
  GETNOS(op3,sp3,tos3,nos3,rol3,fl3,fr3,sr3,ml1,mr1);
  get3:
  op2=(sr2 & 8) ? NOOP : ALU2(op);
  GETNOS(op2,sp2,tos2,nos2,rol2,fl2,fr2,sr2,ml1,mr1);
  op1=(sr1 & 8) ? NOOP : ALU1(op);
  GETNOS(op1,sp1,tos1,nos1,rol1,fl1,fr1,sr1,ml0,mr0);
  op0=(sr0 & 8) ? NOOP : ALU0(op);
  GETNOS(op0,sp0,tos0,nos0,rol0,fl0,fr0,sr0,ml0,mr0);
  nextip=ip;
  break;
  case 2:  if(MASK(op)==0) {
    loops=nextip;
    loope=nextip+DOOFF(op);
  } else
    brip=nextip+BROFF(op);
  goto getall;
  case 3:
    nos3=(cell)nextip;
    brip=(opcode *)((CJLEN(op) || CJABS(op)) ?
		    CFOFF(op) :
		    ((int)nextip)+CFOFF(op));
    if(CJLEN(op)) {
      nos0=nos1=nos2=0;
      op0=op1=op2=NOOP; op3=IP_FETCH; break;
    } else {
      if(CJTYP(op)) {
	nos3=0;
	op3=NOOP;
      } else {
	op3=IP_FETCH;
      }
      goto get3;
    }
  }
  if((nextip==loope) && index) {
    ip=nextip=loops; index--;
  }

/* Execute stack operations */

  ROL(rol0,sp0,tos0);
  ROL(rol1,sp1,tos1);
  ROL(rol2,sp2,tos2);
  ROL(rol3,sp3,tos3);
      
  EXEC(op0,tos0,nos0,sp0,ml0,mr0,sr0,fr0,fl0,fl2);
  EXEC(op1,tos1,nos1,sp1,ml0,mr0,sr1,fr1,fl1,fl3);
  EXEC(op2,tos2,nos2,sp2,ml1,mr1,sr2,fr2,fl0,fl2);
  EXEC(op3,tos3,nos3,sp3,ml1,mr1,sr3,fr3,fl1,fl3);

/* Perform memory and branch operations */

  switch(TYP(op)) {
  case  0:
    if(CU0(op))
      GETEA(dat0,s0e,re,eaevn,imodd, \
	    sp0,sp2,tos0,tos2,sr0,sr2,valevn,numevn,push0)
      else {
	if(dat0 & 0x200)
	  UPDAT(dat0,s0e,re,imodd,valevn,numevn,sp0,sp2,tos0,tos2,sr0,sr2)
push0:    PUSHLD(valevn,numevn,tos0,tos2,sp0,sp2,sr0,sr2);
      }
    if(CU1(op))
      GETEA(dat1,s0o,ro,eaodd,imevn, \
	    sp1,sp3,tos1,tos3,sr1,sr3,valodd,numodd,push1)
      else {
	if(dat1 & 0x200)
	  UPDAT(dat1,s0o,ro,imevn,valodd,numodd,sp1,sp3,tos1,tos3,sr0,sr2)
push1:    PUSHLD(valodd,numodd,tos1,tos3,sp1,sp3,sr1,sr3);
      }
    break;
  case  1: /* conditional setup */
    {
      cell r, cond;

#define GETCOND(sr,tosr,tos,sp) \
      if(EXECQ(sr)) { \
	GETFLAG(r,tos,cond,sr); \
	if(cond & 0x10)  tos=SPT(++sp); \
	tosr = sr | (!r) << 3; \
	} else if((cond & 7) == 0) tosr = sr & ~NO_X;

      cond=COND0(op); GETCOND(sr0,tosr0,tos0,sp0);
      cond=COND1(op); GETCOND(sr1,tosr1,tos1,sp1);
      cond=COND2(op); GETCOND(sr2,tosr2,tos2,sp2);
      cond=COND3(op); GETCOND(sr3,tosr3,tos3,sp3);

      goto pushall;
    }
  case  2: /* branch/do */
    {
      cell cond=COND(op), mask=MASK(op), flag=-(mask >> 4), r;

#define ADDFLAG(sr,tos,sp,maskbit) \
      if((EXECQ(sr)) && (mask & maskbit)) { \
		      GETFLAG(r,tos,cond,sr); \
		      if(cond & 0x10)  tos=SPT(++sp); \
		      if(mask & 0x10)  flag &= r; \
		      else             flag |= r; \
		   }

      ADDFLAG(sr0,tos0,sp0,8);
      ADDFLAG(sr1,tos1,sp1,4);
      ADDFLAG(sr2,tos2,sp2,2);
      ADDFLAG(sr3,tos3,sp3,1);

      if(!flag)  goto pushall;
    }
  case  3:
    if(ip==nextip) ip=nextip=brip;
    else  nextip=brip;
  pushall:
    PUSHLD(valevn,numevn,tos0,tos2,sp0,sp2,sr0,sr2);
    PUSHLD(valodd,numodd,tos1,tos3,sp1,sp3,sr1,sr3);
  }
    if(TYP(op) == 1) {
      sr0 = tosr0;
      sr1 = tosr1;
      sr2 = tosr2;
      sr3 = tosr3;
    }
  }
}
