/* IO-Devices */

#include <stdio.h>
#include "4stack.h"

typedef struct
{
  ucell address;
  ucell range;
  ull (*in_proc) (ucell, ull *);
  void(*out_proc)(ucell, ull, ull *);
} io_proc;

extern io_proc io_list[];

ull io_q;

ull ioin(ucell address, ull * image)
{
  io_proc* list=io_list;
  while(list->range &&
	(address-list->address) >= list->range) list++;
  return (*list->in_proc)(address-list->address, image);
}

void ioout(ucell address, ull data, ull * image)
{   
  io_proc* list=io_list;
  while(list->range &&
	(address-list->address) >= list->range) list++;
  (*list->out_proc)(address-list->address, data, image);
}

ull duart_in(ucell offset, ull * image)
{
  switch(offset) {
  case 0: /* tcx */ return (ull)0;
  case 1: /* tx? */ return (ull)1;
  case 2: /* rcx */
    fflush(stdout);
    {
      int k=0;
      if(key_avail(stdin)) k=getkey(stdin);
/*      fprintf(stderr, "Key %02x '%c' avail\n", k, k); */
      return (ull)k;
    }
  case 3: /* rx? */
    fflush(stdout);
    return (ull)key_avail(stdin);
  default:
    printf("DUART: in %04x\n",offset);
    return 0L;
  }
}

void duart_out(ucell offset, ull data, ull * image)
{
  switch(offset) {
  case 0: /* tcx */ putc(TLO(data), stdout); break;
  case 1: /* tx? */ break;
  case 2: /* rcx */ break;
  case 3: /* rx? */ break;
  default:
    printf("DUART: out %04x,%08x%08x\n",offset,THI(data),TLO(data));
  }
}

extern int w, h, sw, sh, pixels;
extern void *FRAMEBUF;

ull ramdac_in(ucell offset, ull * image)
{
  switch(offset) {
  case 0: return (ull)0;
  case 1: return (ull)(long)FRAMEBUF;
  case 2: return (ull)w;
  case 3: return (ull)sw;
  case 4: return (ull)h;
  case 5: return (ull)sh;
  case 6: return (ull)pixels;
  default:
    printf("RAMDAC: in %04x\n",offset);
    return (ull)0L;
  }
}

void ramdac_out(ucell offset, ull data, ull * image)
{
  /* printf("RAMDAC: out %04x,%08x%08x\n",offset,THI(data),TLO(data)); */

  switch(offset) {
  case 0: /* function register -> store to to activate screen */
    {
      char databuf[10];
      char * args[3];
      
      args[0]="ramdac";
      args[1]="-mag";
      args[2]=databuf;
      
      sprintf(databuf,"%i",(int)data);
      if(!dpy) init(3,args,1);
      else {
	mag=(long)data; resize(); 
      }
      break;
    }
  case 1: /* frame address */
    FRAMEBUF=((void *)image) + (int)data; break;
  case 2: /* screen width */
    w=(int)data; break;
  case 3: /* screen border width */
    sw=(int)data; break;
  case 4: /* screen height */
    h=(int)data; break;
  case 5: /* screen border height */
    sh=(int)data; break;
  case 6: /* sync screen */
    pixels=(int)data; break;
  default: /* nothing happens */
  }
}

void mem_out(ucell offset, ull data, ull* image)
{
  image[offset]=data;
}

ull mem_in(ucell offset, ull* image)
{
  return image[offset];
}

static io_proc io_list[]=
{
  {  -0x100,     0x100, ramdac_in, ramdac_out },
  {  -0x200,     0x020,  duart_in,  duart_out },
  {       0, 0x0200000,    mem_in,    mem_out },
  {       0,         0,         0,          0 }
};
