/* stuff for MSDOS virtual screen display */

#include <graphics.h>
#include <stdio.h>

extern int LINES;

#ifdef EQUAL_COLOR_DISTANCE
/* dither values */
#   define RDITHER 2 /* read is 32% sensitive */
#   define GDITHER 2 /* green is 66% sensitive */
#   define BDITHER 2 /* and blue is the rest 2% sensitive */
#else
/* optional dither parameter */
#   define RDITHER 2 /* read is 32% sensitive */
#   define GDITHER 4 /* green is 66% sensitive */
#   define BDITHER 2 /* and blue is the rest 2% sensitive */
#endif
/* gives a total amout of colors: */
#define PALETTE RDITHER*GDITHER*BDITHER
#define PIXREC 16
#define PIXNUM 256

unsigned char *screenimg=0xD0200000;
unsigned char *screenmap;
int screen;
static int color_offset=0;

extern int *FRAMEBUF;
int dpy=0;

int w=0, h=0;
int mag=1;
int sw=640, sh=480;

#define fatal_error(arg)  {fprintf(stderr,arg); exit(1);}

void update_screen(void)
{
   int i, j, k;
   int planelen=(w*h*mag*mag)/8;
   char *frame=screenmap;
   char *screen;
   char image0[planelen];
   char image1[planelen];
   char image2[planelen];
   char image3[planelen];
   char* img0=image0;
   char* img1=image1;
   char* img2=image2;
   char* img3=image3;
   int slen = sw/8;
   int plen = w*mag/8;
   unsigned char p0=0, p1=0, p2=0, p3=0;
   unsigned char accu;
   
   for(i=0; i<h*mag; i++)
     for(j=0; j<plen; j++)
     {
	for(k=0; k<8; k++)
	  {
	     accu = *frame++;
	     
	     p0 = (p0 << 1) | (accu & 1); accu >>= 1;
	     p1 = (p1 << 1) | (accu & 1); accu >>= 1;
	     p2 = (p2 << 1) | (accu & 1); accu >>= 1;
	     p3 = (p3 << 1) | (accu & 1); accu >>= 1;
	  }
	*img0++ = p0;
	*img1++ = p1;
	*img2++ = p2;
	*img3++ = p3;
     }

   img0=image0;
   img1=image1;
   img2=image2;
   img3=image3;

   screen=screenimg;
   for(i=h*mag; --i>=0;)
    {
      memcpy(screen,img0,plen);
      img0+=plen; screen+=slen;
    }
   
   screen=screenimg+0x10000;
   for(i=h*mag; --i>=0;)
     {
	memcpy(screen,img1,plen);
	img1+=plen; screen+=slen;
     }
   
   screen=screenimg+0x20000;
   for(i=h*mag; --i>=0;)
     {
	memcpy(screen,img2,plen);
	img2+=plen; screen+=slen;
     }
   
   screen=screenimg+0x30000;
   for(i=h*mag; --i>=0;)
     {
	memcpy(screen,img3,plen);
	img3+=plen; screen+=slen;
     }
   
}

void end_graphics(void)
{
}

#define COLORS 0xa8

static void load_palette(void)
{
   int i=0, r, g, b;

   for(g=0; g<GDITHER; g++)
      for(r=0; r<RDITHER; r++)
         for(b=0; b<BDITHER; b++, i++)
            {
	       GrSetColor(((i < 8) ? i : i+0x30),
			  (r*COLORS)/(RDITHER-1),
			  (g*COLORS)/(GDITHER-1),
			  (b*COLORS)/(BDITHER-1));
	    }
}

void init(int argc, char** argv)
{
  dpy=1;
  mag=1;

  while (++argv, --argc)
    {
      if (!strcmp(*argv,"-wh") && argc >= 2)
	{
	  w=atoi(*(++argv));argc--;
	  h=atoi(*(++argv));argc--;
	}
      else if (!strcmp(*argv,"-swh") && argc >= 2)
	{
	  sw=atoi(*(++argv));argc--;
	  sh=atoi(*(++argv));argc--;
	}
      else if (!strcmp(*argv,"-mag") && argc >= 1)
	{
	  mag=atoi(*(++argv));argc--;
	}
      else fprintf(stderr,"Unrecognised option: %s\n",*argv);
    }
  
  if(!w) w=sw;
  if(!h) h=sh;

  GrSetMode(6,sw,sh);
  LINES=sh/16;

  load_palette();

  w = ((w-1) & -PIXREC) + PIXREC;
  h = ((h-1) & -PIXREC) + PIXREC;

  screenmap = malloc(h*w*mag*mag);
  bzero(screenmap,h*w*mag*mag);
}

void resize()
{
  GrSetMode(6,sw,sh);
  LINES=sh/16;

  w = ((w-1) & -PIXREC) + PIXREC;
  h = ((h-1) & -PIXREC) + PIXREC;

  free(screenmap);
  screenmap = malloc(h*w*mag*mag);
  bzero(screenmap,h*w*mag*mag);
}


typedef union
{
  struct
    {
#ifdef BIG_ENDIAN
      unsigned char alpha, r, g, b;
#else
      unsigned char b, g, r, alpha;
#endif
    }
  rgb;
  int col;
}
color;

void dither(int *frame, char *image, int w, int h)
{
  int r=0, g=0, b=0;
  color ditherbuf[PIXNUM];
  unsigned char framebuf[PIXNUM];
  int i, j, k, p, q;
  int coladd=color_offset;
  static int hilbert_mask[PIXNUM]={
#include "hilbert.h"
  };

  for(i=0; i<h; i+=PIXREC/mag)
    for(j=0; j<w; j+=PIXREC/mag)
      {
	int *framei=frame+i*w;
#ifdef BYTE_ALIGNED_LINES
	char *imagei=image+i*w*mag*mag;
#else
        int *imagei=(int *)(image+i*w*mag*mag);
#endif
	k=0;
	
	if(mag==1)
	  for(p=i; p<i+PIXREC; p++, framei+=w)
	    for(q=j; q<j+PIXREC; q++)
	      ditherbuf[k++].col=framei[q^1];
	else
	  for(p=i;p<i+PIXREC/mag; p++, framei+=w)
	    {
	      int m,n;
	      
	      for(m=0; m<mag; m++)
	        for(q=j; q<j+PIXREC/mag; q++)
	          for(n=0; n<mag; n++)
	            ditherbuf[k++].col=framei[q^1];
	    }
	
	for(k=0; k<PIXNUM; k++)
	  {
	    unsigned int pixcol;
	    color pixel;
	    int hilbert;
	     
	    hilbert=hilbert_mask[k];

	    pixel=ditherbuf[hilbert].col;
	    r+=(((int)pixel.rgb.r) * (((RDITHER-1) << 8) + RDITHER)) >> 8;
	    g+=(((int)pixel.rgb.g) * (((GDITHER-1) << 8) + GDITHER)) >> 8;
	    b+=(((int)pixel.rgb.b) * (((BDITHER-1) << 8) + BDITHER)) >> 8;

	    pixcol = ((g >> 8) * RDITHER + (r >> 8)) * BDITHER + (b >> 8);

	    framebuf[hilbert]=(unsigned char)(pixcol + coladd);

	    r &= 0xff;
	    g &= 0xff;
	    b &= 0xff;
	  }

#ifdef BYTE_ALIGNED_LINES
	for(p=i*mag, k=0; p<i*mag+PIXREC; p++, imagei+=w*mag)
	  for(q=j*mag; q<j*mag+PIXREC; q++, k++)
	    imagei[q]=framebuf[k];
#else
	 /* assumes sizeof(int)==4 */
	  {
	     int *frame=(int *)framebuf;
	     int jj=j*mag/sizeof(int), ww=w*mag/sizeof(int);
	     
	     for(p=0; p<PIXREC; p++, imagei+=ww)
	       {
		  imagei[jj+0]=*frame++;
		  imagei[jj+1]=*frame++;
		  imagei[jj+2]=*frame++;
		  imagei[jj+3]=*frame++;
	       }
	  }
#endif
      }
}

void check_event(int what_to_do)
{
  if((what_to_do & 2) && !kbhit())
    {
      dither(FRAMEBUF, screenmap, w, h);
      update_screen();
    }
}

