/* stuff for X Window virtual screen display */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#ifndef NO_XSHM
#   include <sys/shm.h>
#   include <X11/extensions/XShm.h>
#endif
#include <stdio.h>
#include <stdlib.h>

#if defined(i386) || defined(i486) || defined(i586)
#   define BIGENDIAN 0
#else
#   define BIGENDIAN 1
#endif

#define setink(inkcolor) XSetForeground (dpy,gc,inkcolor)

#define white WhitePixel(dpy, screen)
#define black BlackPixel(dpy, screen)

#ifdef EQUAL_COLOR_DISTANCE
/* dither values */
#   define RDITHER 6 /* read is 32% sensitive */
#   define GDITHER 6 /* green is 66% sensitive */
#   define BDITHER 6 /* and blue is the rest 2% sensitive */
#else
/* optional dither parameter */
#   define RDITHER 6  /* read is 32% sensitive */
#   define GDITHER 10 /* green is 66% sensitive */
#   define BDITHER 4  /* and blue is the rest 2% sensitive */
#endif
/* gives a total amout of colors: */
#define PALETTE RDITHER*GDITHER*BDITHER
#define PIXREC 16
#define PIXNUM 256

GC	gc;
static int	event_mask= (ExposureMask
			     | StructureNotifyMask
			     | ButtonPressMask
			     | KeyPressMask
			     | PointerMotionMask);
Window	win;
Display	*dpy=0;
XImage	*screenimg;
unsigned char *screenmap;
int screen;
int pixels;

extern int *FRAMEBUF;

int linew;
int mag=1;
int w=0, h=0;
int sw, sh; /* dummies, there is no border on X windows */

static XVisualInfo *visual_info_list; 
static Colormap colormap;

static int color_offset=0;

#ifndef NO_XSHM
static XShmSegmentInfo xshminfo;
#endif

#define fatal_error(arg)  {fprintf(stderr,arg); exit(1);}

int depth, bitppix;

static void get_visual(void)
{
  XVisualInfo desired_visual_template;
  int no_visual_matched, n, i;
  XPixmapFormatValues * pmformat;

  desired_visual_template.screen = screen;
  visual_info_list = XGetVisualInfo(dpy, 
				    VisualScreenMask /*| VisualClassMask*/,
				    &desired_visual_template,
				    &no_visual_matched);
  pmformat = XListPixmapFormats(dpy, &n);
  depth = 0;
  bitppix = 0;
  for(i=0; i<n; i++) {
    if(pmformat[i].depth == 32 || pmformat[i].depth == 24) {
      depth = pmformat[i].depth;
      if(bitppix != 32)
	bitppix = pmformat[i].bits_per_pixel;
    } else if(pmformat[i].depth == 16 &&
	      (bitppix != 32 && bitppix != 24)) {
      depth = pmformat[i].depth;
      bitppix = pmformat[i].bits_per_pixel;
    } else if(pmformat[i].depth == 8 &&
	      (bitppix != 32 && bitppix != 24 && bitppix != 16)) {
      depth = pmformat[i].depth;
      bitppix = pmformat[i].bits_per_pixel;
    }
  }
}

#ifdef NO_XSHM
void map_shared_mem(void)
{
  XWindowAttributes win_attributes;
  
  screenimg = XCreateImage(dpy, visual_info_list[0].visual, depth,
			   ZPixmap, 0,screenmap, w*mag, h*mag, 8, 0);
}

void update_screen(char *fb)
{
  XPutImage(dpy, win, gc, screenimg, 0, 0, 0, 0, w*mag, h*mag);
  XSync(dpy, 0);
}


void end_graphics(void)
{
}

#else
static int shmem=1;

static void map_shared_mem(void)
{
  if(!shmem) {
    screenimg = XCreateImage(dpy, visual_info_list[0].visual, depth,
			     ZPixmap, 0, screenmap, w*mag, h*mag, 8, 0);
    return;
  }

  screenimg = XShmCreateImage(dpy, visual_info_list[0].visual, depth,
			      ZPixmap, screenmap, &xshminfo,
			      w*mag, h*mag);
  xshminfo.shmid = shmget(IPC_PRIVATE,
			  screenimg->bytes_per_line * screenimg->height,
			  IPC_CREAT | 0777);
  if (xshminfo.shmid < 0)
    fatal_error("shmget failed");
  
  screenimg->data = (unsigned char *) shmat(xshminfo.shmid, 0, 0);
  if (screenimg->data == ((char *) -1))
    fatal_error("shmat failed");
  
  screenmap = screenimg->data;
  xshminfo.shmaddr = screenimg->data;
  xshminfo.readOnly = True;
  if (!XShmAttach(dpy, &xshminfo))
    fatal_error("Unable to map shared memory.\n");
}


void update_screen(char *fb)
{
  if(shmem)
    XShmPutImage(dpy, win, gc, screenimg, 0, 0, 0, 0, w*mag, h*mag, 0);
  else
    XPutImage(dpy, win, gc, screenimg, 0, 0, 0, 0, w*mag, h*mag);
  XSync(dpy, 0);
}


void end_graphics(void)
{
  if(!shmem) return;
  XShmDetach(dpy, &xshminfo);
  XDestroyImage(screenimg);
  shmdt(xshminfo.shmaddr);
  shmctl(xshminfo.shmid, IPC_RMID, 0);
}
#endif

static void load_palette(void)
{
  int i, r, g, b;
  Colormap oldcolors;
  XColor palette[PIXNUM];

  oldcolors = DefaultColormap (dpy, screen);
  color_offset= (1 << DefaultDepth(dpy, screen)) -
    RDITHER * GDITHER * BDITHER;

  /*   if((XAllocColorCells (dpy, colormap, 0, &dummy, 1, palette, PALETTE)))
       color_offset = palette[0].pixel;
       else */
  colormap = XCreateColormap(dpy, DefaultRootWindow(dpy),
			     visual_info_list[0].visual, AllocAll);

  for(i=0; i<PIXNUM; i++) {
    palette[i].pixel=i;
  }
  XQueryColors(dpy, oldcolors, palette, color_offset);

  i=color_offset;

  for(r=0; r<RDITHER; r++)
    for(g=0; g<GDITHER; g++)
      for(b=0; b<BDITHER; b++, i++) {
	palette[i].pixel = i;
	palette[i].red   = (int)((r * 65535)/(RDITHER-1));
	palette[i].green = (int)((g * 65535)/(GDITHER-1));
	palette[i].blue  = (int)((b * 65535)/(BDITHER-1));
	palette[i].flags = DoRed | DoGreen | DoBlue;
	/*	     printf("Color %i: r:%04x, g:%04x, b:%04x\n",i,
		     palette[i].red,
		     palette[i].green,
		     palette[i].blue);
		     */
      }

  XStoreColors(dpy, colormap, palette, PIXNUM);
}

void init(int argc, char** argv, int flag)
{
  XGCValues *xgc;
  static char* display = NULL;
  static char* geometry = NULL;
  char* host = getenv("HOST");
  XSetWindowAttributes xswa;
  XSizeHints hints;
  int xswamask=0, retval=0;
  int x=0, y=0;
  XEvent event;
  static int nodpy=0;

  if(!display) display = getenv("DISPLAY");
  if(!geometry) geometry = "";

  mag=1;
  pixels=-1;

  while (++argv, --argc) {
    if (!strcmp(*argv,"-display") && argc >= 1) {
      display=*(++argv); argc--;
    } else if (!strcmp(*argv,"-geometry") && argc >= 1) {
      geometry=*(++argv); argc--;
    } else if (!strcmp(*argv,"-mag") && argc >= 1) {
      mag=atoi(*(++argv)); argc--;
    } else if (!strcmp(*argv,"-nodpy")) {
      nodpy = 1;
    } else fprintf(stderr,"Unrecognised option: %s\n",*argv);
  }

  if(!flag || nodpy) return;

  xgc=( XGCValues *) malloc(sizeof(XGCValues) );
  
  if(!display[0]) fatal_error("Display variable not set!");

  shmem = (display[0] == ':') ||
          (!memcmp(host,display,strlen(host)) &&
           display[strlen(host)]==':');
  
  if (!(dpy = XOpenDisplay (display))) {
    fprintf(stderr,"Error:  Can't open display '%s'!\n",display);
    exit(1);
  }
  
  screen = DefaultScreen (dpy);
  get_visual();
  if(depth==8)
    load_palette();
  if(!w) w=DisplayWidth (dpy, screen)/mag;
  if(!h) h=DisplayHeight (dpy, screen)/mag;
  
  if(geometry[0])
    retval=XParseGeometry(geometry,&x,&y,&w,&h);

  w = ((w-1) & -PIXREC) + PIXREC;
  h = ((h-1) & -PIXREC) + PIXREC;
  pixels = w*h;

  if (retval & XValue && retval & XNegative)
    x = XDisplayWidth(dpy,screen)-w-abs(x);
  if (retval & YValue && retval & YNegative)
    y = XDisplayHeight(dpy,screen)-h-abs(y);
  hints.flags = 0;
  if(retval & (XValue | YValue))
    hints.flags |= USPosition;
  hints.flags |= USSize | PResizeInc | PBaseSize;
  hints.x = x; hints.y = y;
  hints.width = w*mag; hints.height = h*mag;
  hints.base_width = hints.base_height = 0;
  hints.width_inc = hints.height_inc = PIXREC*mag;
  
  xswamask |= CWBackPixel | CWBorderPixel | CWColormap;
  xswa.background_pixel=BlackPixel(dpy, screen);
  xswa.border_pixel=BlackPixel(dpy, screen);
  xswa.colormap=colormap;
  
  win = XCreateWindow (dpy, DefaultRootWindow (dpy),
		       x, y, w*mag, h*mag, 1, 0,
		       CopyFromParent, CopyFromParent,
		       xswamask, &xswa);
  
  XSetStandardProperties(dpy,win,"4stack virtual screen","4stack", None,
			 0,0,&hints);
  XMapWindow (dpy, win);
  XSelectInput (dpy, win, event_mask);
  
/*  xswa.override_redirect = 0;*/
  
  gc = XCreateGC (dpy, win, 0, xgc);
  
  screenmap = (char *) malloc(h*w*mag*mag*(bitppix >> 3));
  bzero(screenmap,h*w*mag*mag*(bitppix >> 3));
  
  map_shared_mem();

  do {
    XNextEvent(dpy, &event);
  } while (event.type != Expose);
  
  setink(black);
  XFillRectangle (dpy, win, gc, 0, 0, w, h);
  
  XSync(dpy,0);
}

void resize()
{
  w = ((w-1) & -PIXREC) + PIXREC;
  h = ((h-1) & -PIXREC) + PIXREC;

  /* to define!!! Resize window!!! */

  free(screenmap);
  screenmap = (char *) malloc(h*w*mag*mag*(bitppix >> 3));
  bzero(screenmap, h*w*mag*mag*(bitppix >> 3));
}

typedef union
{
  struct
    {
#if BIGENDIAN
      unsigned char alpha, r, g, b;
#else
      unsigned char b, g, r, alpha;
#endif
    }
  rgb;
  int col;
}
color;

#if BIGENDIAN
#  define FB(i) (i)
#else
#  define FB(i) ((i)^1)
#endif

void dither(int *frame, char *image, int w, int h)
{
  int r=0, g=0, b=0;
  color ditherbuf[PIXNUM];
  unsigned char framebuf[PIXNUM];
  int i, j, k, p, q;
  int coladd=color_offset;
  static int hilbert_mask[PIXNUM]={
#include "hilbert.h"
  };

/*  fprintf(stderr, "depth=%d, rgb=%d\n", depth, bitppix); */

  if(bitppix == 32) {
    int *framei=frame;
    char *imagei=image;
    color pixel;
    int k, l;

    if(mag == 1) {
      for(i=0; i<h; i++) {
	color p1, p2;
	for(j=0; j<w; j+=2) {
	  p1 = (color)framei[0];
	  p2 = (color)framei[1];
	  ((color*)imagei)[!BIGENDIAN] = p1;
	  ((color*)imagei)[ BIGENDIAN] = p2;
	  imagei += 8;
	  framei += 2;
	}
      }
    } else {
      for(i=0; i<h; i++) {
	for(l=0; l<mag; l++) {
	  for(j=0; j<w; j++) {
	    pixel = *(color *)(((int)framei) ^ (!BIGENDIAN << 2));
	    for(k=0; k<mag; k++) { 
	      *((color*)imagei)++ = pixel;
	    }
	    framei++;
	  }
	  framei -= w;
	}
	framei += w;
      }
    }
    return;
  } else if(bitppix == 24) {
    int *framei=frame;
    char *imagei=image;
    color pixel;
    int k, l;
    
    if(mag == 1) {
      for(i=0; i<h; i++) {
	color p1, p2;
	for(j=0; j<w; j+=2) {
	  p1 = (color)framei[0];
	  p2 = (color)framei[1];
	  if(BIGENDIAN) {
	    imagei[0]=p1.rgb.r;
	    imagei[1]=p1.rgb.g;
	    imagei[2]=p1.rgb.b;
	    imagei[3]=p2.rgb.r;
	    imagei[4]=p2.rgb.g;
	    imagei[5]=p2.rgb.b;
	  } else {
	    imagei[0]=p2.rgb.b;
	    imagei[1]=p2.rgb.g;
	    imagei[2]=p2.rgb.r;
	    imagei[3]=p1.rgb.b;
	    imagei[4]=p1.rgb.g;
	    imagei[5]=p1.rgb.r;
	  }
	  imagei += 6;
	  framei += 2;
	}
      }
    } else {
      for(i=0; i<h; i++) {
	for(l=0; l<mag; l++) {
	  for(j=0; j<w; j++) {
	    pixel = *(color *)(((int)framei) ^ (!BIGENDIAN << 2));
	    for(k=0; k<mag; k++) {
	      if(BIGENDIAN) {
		*(imagei)++ = pixel.rgb.r;
		*(imagei)++ = pixel.rgb.g;
		*(imagei)++ = pixel.rgb.b;
	      } else {
		*(imagei)++ = pixel.rgb.b;
		*(imagei)++ = pixel.rgb.g;
		*(imagei)++ = pixel.rgb.r;
	      }
	    }
	    framei++;
	  }
	  framei -= w;
	}
	framei += w;
      }
    }
    return;
  } else if(bitppix == 16) {
    int *framei=frame;
    char *imagei=image;
    color pixel;
    int k, l;
    static int init=0;
    static short reds[256+8], greens[256+4], blues[256+8];
    
    if(!init) {
      for(i=0; i<256+8; i++)
	reds  [i] = ((i<<8)/(256+7) & 0xf8) << 8;
      for(i=0; i<256+4; i++)
	greens[i] = ((i<<8)/(256+3) & 0xfc) << 3;
      for(i=0; i<256+8; i++)
	blues [i] = ((i<<8)/(256+7) & 0xf8) >> 3;
      init=1;
    }

    if(mag == 1) {
      for(i=0; i<h; i++) {
	color p1, p2;
	int go0=(h&1) ? 1 : 0;
	int go1=(h&1) ? 3 : 2;
	int ro0=go0*2;
	int ro1=go1*2;

	for(j=0; j<w; j+=2) {
	  p1 = (color)framei[0];
	  p2 = (color)framei[1];
	  if(BIGENDIAN) {
	    ((short*)imagei)[0]=(reds  [p1.rgb.r+ro0] |
				 greens[p1.rgb.g+go0] |
				 blues [p1.rgb.b+ro0]);
	    ((short*)imagei)[1]=(reds  [p2.rgb.r+ro1] |
				 greens[p2.rgb.g+go1] |
				 blues [p2.rgb.b+ro1]);
	  } else {
	    ((short*)imagei)[0]=(reds  [p2.rgb.r+ro1] |
				 greens[p2.rgb.g+go1] |
				 blues [p2.rgb.b+ro1]);
	    ((short*)imagei)[1]=(reds  [p1.rgb.r+ro0] |
				 greens[p1.rgb.g+go0] |
				 blues [p1.rgb.b+ro0]);
	  }
	  imagei += 4;
	  framei += 2;
	}
      }
    } else {
      int ditherh, ditherv=0;
      for(i=0; i<h; i++) {
	for(l=0; l<mag; l++) {
	  int go0=(ditherv&1) ? 0 : 1;
	  int go1=(ditherv&1) ? 2 : 3;
	  ditherv++;
	  ditherh=0;

	  for(j=0; j<w; j++) {
	    pixel = *(color *)(((int)framei) ^ (!BIGENDIAN << 2));
	    for(k=0; k<mag; k++) {
	      int go=((ditherh^ditherv)&1) ? go1 : go0;
	      int ro=go*2;
	      ditherh++;

	      *((short*)imagei)++=(reds  [pixel.rgb.r+ro] |
				   greens[pixel.rgb.g+go] |
				   blues [pixel.rgb.b+ro]);
	    }
	    framei++;
	  }
	  framei -= w;
	}
	framei += w;
      }
    }
    return;
  }

  for(i=0; i<h; i+=PIXREC/mag)
    for(j=0; j<w; j+=PIXREC/mag) {
      int *framei=frame+i*w;
#ifdef BYTE_ALIGNED_LINES
      char *imagei=image+i*w*mag*mag;
#else
      int *imagei=(int *)(image+i*w*mag*mag);
#endif
      k=0;
      
      if(mag==1)
	for(p=i; p<i+PIXREC; p++, framei+=w)
	  for(q=j; q<j+PIXREC; q++)
	    ditherbuf[k++].col=framei[FB(q)];
      else
	for(p=i;p<i+PIXREC/mag; p++, framei+=w) {
	  int m,n;
	  
	  for(m=0; m<mag; m++)
	    for(q=j; q<j+PIXREC/mag; q++)
	      for(n=0; n<mag; n++)
		ditherbuf[k++].col=framei[FB(q)];
	}
      
      for(k=0; k<PIXNUM; k++) {
	unsigned int pixcol;
	color pixel;
	int hilbert;
	
	hilbert=hilbert_mask[k];
	
	pixel=(color)ditherbuf[hilbert].col;
	r+=(((int)pixel.rgb.r) * (((RDITHER-1) << 8) + RDITHER)) >> 8;
	g+=(((int)pixel.rgb.g) * (((GDITHER-1) << 8) + GDITHER)) >> 8;
	b+=(((int)pixel.rgb.b) * (((BDITHER-1) << 8) + BDITHER)) >> 8;
	
	pixcol = ((r >> 8) * GDITHER + (g >> 8)) * BDITHER + (b >> 8);
	
	framebuf[hilbert]=(unsigned char)(pixcol + coladd);
	
	r &= 0xff;
	g &= 0xff;
	b &= 0xff;
      }

#ifdef BYTE_ALIGNED_LINES
      for(p=i*mag, k=0; p<i*mag+PIXREC; p++, imagei+=w*mag) {
	memmove(imagei[j*mag], framebuf[k], PIXREC);
	k+=PIXREC;
      }
#else
	 /* assumes sizeof(int)==4 */
      {
	int *frame=(int *)framebuf;
	int jj=j*mag/sizeof(int), ww=w*mag/sizeof(int);
	
	for(p=0; p<PIXREC; p++, imagei+=ww) {
	  imagei[jj+0]=*frame++;
	  imagei[jj+1]=*frame++;
	  imagei[jj+2]=*frame++;
	  imagei[jj+3]=*frame++;
	}
      }
#endif
    }
  }

void save_image(int *frame, int w, int h)
{
  static int picnum=0;
  char filename[20];
  FILE *picfile;

  char picline[w*3];
  int *framep=frame;
  color col;
  char *picp;
  int i, j, k=0;

  sprintf(filename,"./frame%i.ppm",picnum++);

  picfile=fopen(filename,"wb");

  fprintf(picfile,"P6\n# CREATOR: 4stack 0.0\n%i %i\n255\n",w,h);

  for(i=0; i<h; i++)
    {
      picp=picline;
      for(j=0; j<w; j++)
	{
	  col.col=framep[FB(k++)];
	  *picp++ = col.rgb.r;
	  *picp++ = col.rgb.g;
	  *picp++ = col.rgb.b;
	}
      fwrite(picline, 1, 3*w, picfile);
    }
  fclose(picfile);
}

void check_event(int what_to_do)
{
  XEvent event;
  struct timeval timeout;
  
  while(XEventsQueued (dpy, QueuedAfterReading)) {
    XWindowEvent (dpy, win, event_mask, &event);
    
    switch (event.type) {
    case ConfigureNotify:
      if (event.xconfigure.width != w*mag ||
	  event.xconfigure.height != h*mag ) {
	/* fprintf(stderr,"WxH: %ix%i\n",w,h); */
	w = event.xconfigure.width/mag;
	h = event.xconfigure.height/mag;
	/* fprintf(stderr,"WxH: %ix%i\n",w,h); */
	
	w = ((w-1) & -PIXREC) + PIXREC;
	h = ((h-1) & -PIXREC) + PIXREC;
	
	end_graphics();
	free(screenmap);
	screenmap = (char *) malloc(w*h*mag*mag);
	map_shared_mem();
	dither(FRAMEBUF, screenmap, w, h);
	update_screen(screenmap);
      }
    case Expose:
	update_screen(screenmap);
    case KeyPress:
      {
	KeySym kpr=XKeycodeToKeysym(dpy,event.xkey.keycode,0);
	if (kpr=='s') /* s */
	  save_image(FRAMEBUF, w, h);
      }
    default: break;
    }
  }
  
  if(what_to_do & 2) {
    dither(FRAMEBUF, screenmap, w, h);
    update_screen(screenmap);
  }
  if(what_to_do & 1) {
    timeout.tv_sec=0;
    timeout.tv_usec=40000;
    (void)select(0,0,0,0,&timeout);
  }
}

